"""

Create an EOS class from options taken from factory design pattern.

"""

# Add imports here
from importlib import import_module
import logging

class jit_stat:
    disable_jit = True

def eos(**kwargs):
    """
    Interface between the user and our library of equations of state (EOS).

    Input the name of a desired EOS and a factory design pattern with a dictionary will search available classes to allow easy implementation of new EOS.

    Parameters
    ----------
        kwargs : dict, Optional
            A dictionary of inputs for the desired EOS. See specific EOS documentation for required inputs.

            - eos : str - Input should be in the form EOSfamily.EOSname (e.g. saft.gamme_mie). Note that the name of the class is EOSfamily_EOSname.
                
    Returns
    -------
        instance : obj
            An instance of the defined EOS class to be used in thermodynamic computations.
    """

    logger = logging.getLogger(__name__)

    if "eos" not in kwargs:
        eos_type = "saft.gamma_mie"
        logger.info("Trying default EOS, {}".format(eos_type))
    else:
        eos_type = kwargs["eos"]  
        del kwargs["eos"]
        logger.info("Trying user defined EOS, {}".format(eos_type))

    if 'jit' not in kwargs:
        jit_stat.disable_jit = True
    else:
        jit_stat.disable_jit = not kwargs['jit']

    try:
        eos_fam, eos = eos_type.split('.')
    except:
        raise Exception("Input should be in the form EOSfamily.EOSname (e.g. saft.gamme_mie).")

    try:
        eos_module = import_module('.' + eos, package="despasito.equations_of_state." + eos_fam)
        class_name = "_".join([eos_fam, eos])
        eos_class = getattr(eos_module, class_name)
        instance = eos_class(kwargs)
    except (AttributeError):
        raise ImportError(
            "Based on your input, '{}', we expect the class, {}, in a module, {}, found in the package, {}, which indicates the EOS family.".format(eos_type, class_name, eos, eos_fam))
    logger.info("Created {} eos object".format(eos_type))

    return instance




"""

Routines for parsing input .json files to dictionaries for program use, and write .json files.

"""

import logging
import json
import collections
import numpy as np
import os

######################################################################
#                                                                    #
#                   Appends path to find data files                  #
#                                                                    #
######################################################################

def append_data_file_path(input_dict, path='.'):
   """
   Appends path to data file(s).

   Parameters
   ----------
   input_dict
       Dictionary of input (json) data 
   path: str
       relative path to append to existing data files 
   """
   if 'EOSgroup' in input_dict:
      input_dict['EOSgroup'] = os.path.join(path, input_dict['EOSgroup'])

   if 'EOScross' in input_dict:
      input_dict['EOScross'] = os.path.join(path, input_dict['EOScross'])

   for key, val in input_dict.items():
      if 'file' in val:
         input_dict[key]['file'] = os.path.join(path, input_dict[key]['file'])

######################################################################
#                                                                    #
#                  Extract Bead Data                                 #
#                                                                    #
######################################################################
def extract_calc_data(input_fname, path='.', **args):

    """
    Parse dictionary from .json input file into a dictionaries.

    Resulting dictionaries are used for creating the equation of state object, and for passing instructions for thermodynamic calculations.

    Parameters
    ----------
    input_fname : str
        The file name of a .json file in the current directory containing (1) the paths to equation of state parameters, (2) :mod:`~despasito.thermodynamics.calc_types` and inputs for thermodynamic calculations (e.g. density options for :func:`~despasito.thermodynamics.calc.PvsRho`).

    Returns
    -------
    eos_dict : dict
        Dictionary of bead definitions and parameters used to later initialize eos object. :func:`despasito.equations_of_state.eos`
    thermo_dict : dict
        Dictionary of instructions for thermodynamic calculations or parameter fitting. :func:`despasito.thermodynamics.thermo`
    """

    logger = logging.getLogger(__name__)

    ## Extract dictionary from input file
    with open(input_fname, 'r') as input_file:
        input_dict = json.loads(input_file.read())

    # Look for data (library) files in the user-supplied path
    append_data_file_path(input_dict, path)

    if "output_file" in input_dict:
        output_file = input_dict["output_file"]
    else:
        output_file = None

    ## Make bead data dictionary for EOS
    #process input file
    beads, nui = process_bead_data(input_dict['beadconfig'])
    eos_dict = {'beads':beads,'nui':nui}

    #read EOS groups file
    with open(input_dict['EOSgroup'], 'r') as f:
        output = f.read()
    eos_dict['beadlibrary'] = json.loads(output)

    #read EOS cross file
    try:
        with open(input_dict['EOScross'], 'r') as f:
            output = f.read()
        eos_dict['crosslibrary'] = json.loads(output)
        logger.info("Cross interaction parameters have been accepted")
    except:
        logger.info("No EOScross file specified")

    try:
        eos_dict['sitenames'] = input_dict['association_site_names']
        logger.info('Association sites have been accepted')
    except:
        logger.info('No association sites specified')

    if "eos" in input_dict:
        eos_dict['eos'] = input_dict["eos"]

    ## Make dictionary of data needed for thermodynamic calculation
    thermo_dict = {}
    # Extract relevant system state inputs
    EOS_dict_keys = ['beadconfig', 'EOSgroup', 'EOScross','association_site_names',"output_file","eos"]
    for key, value in input_dict.items():
        if key not in EOS_dict_keys:
            thermo_dict[key] = value

    if "opt_params" not in thermo_dict:
        logger.info("The following thermo calculation parameters have been provided: {}\n".format((", ".join(thermo_dict.keys()))))
    else: # parameter fitting
        thermo_dict = process_param_fit_inputs(thermo_dict)
        tmp = ""
        for key, value in thermo_dict["exp_data"].items():
            tmp += " {} ({}),".format(key,value["name"])
        logger.info("The bead, {}, will have the parameters {}, fit using the following data:\n {}".format(thermo_dict["opt_params"]["fit_bead"],thermo_dict["opt_params"]["fit_params"],tmp))

    return eos_dict, thermo_dict, output_file

######################################################################
#                                                                    #
#                  Extract Density Plot  Parameters                  #
#                                                                    #
######################################################################
def file2paramdict(filename,delimiter=" "):

    """
    Converted file directly into a dictionary.

    Each line in the input file is a key followed by a value in the resulting dictionary.

    Parameters
    ----------
    filename : str
        File of keys and values
    delimiter : str, Optional, default: " "
        String separating key and value within file

    Returns
    -------
    dictionary : dict
        Resulting dictionary
    """

    #logger = logging.getLogger(__name__)

    dictionary = {}
    with  open(filename, "r") as filedata:
        for line in filedata:
            line.rstrip()
            linearray = line.split(delimiter)
            if len(linearray) == 2:
                try:
                    dictionary[linearray[0]] = eval(linearray[1])
                except:
                    if line[1].isdigit():
                        dictionary[linearray[0]] = float(linearray[1])
                    else:
                        dictionary[linearray[0]] = linearray[1]
    
    return dictionary

######################################################################
#                                                                    #
#                  Extract Saft Parameters                           #
#                                                                    #
######################################################################
def write_EOSparameters(library, filename):

    """
    Sort and export dictionary of input parameters into .json file.

    Parameters
    ----------
    library : dict
        Dictionary of parameters to be sorted and exported 

    Returns
    -------
    filename : str
        Filename (with or without path) of .json file of parameters
    """

    #logger = logging.getLogger(__name__)

    #sort and write SAFT dict
    for i in library:
        library[i] = collections.OrderedDict(sorted(list(library[i].items()), key=lambda tup: tup[0].lower()))
    f = open(filename, 'w')
    json.dump(library, f, indent=4)


######################################################################
#                                                                    #
#                  Make Mole Frac. Matrix                            #
#                                                                    #
######################################################################
def make_xi_matrix(filename):

    """
    Extract system information from input .json file of parameters.

    Parameters
    ----------
    filename : str
        Name of .json file containing system and thermodynamic information. This function deals with the system information.

    Returns
    -------
    xi : list[float]
        Mole fraction of component, only relevant for parameter fitting
    beads : list[str]
        List of unique bead names used among components
    nui : numpy.ndarray
        Array of number of components by number of bead types. Defines the number of each type of group in each component.
    """

    #logger = logging.getLogger(__name__)

    f = open(filename, 'r').read()
    comp = json.loads(f)
    beads, nui = process_bead_data(comp)
    return xi, beads, nui


######################################################################
#                                                                    #
#                  Process Bead Data                                 #
#                                                                    #
######################################################################
def process_bead_data(bead_data):

    """
    Process system information from input file.

    Parameters
    ----------
    bead_data : list[list]
        List of strings and dictionaries from .json file

    Returns
    -------
    beads : list[str]
        List of unique bead names used among components
    nui : numpy.ndarray
        Array of number of components by number of bead types. Defines the number of each type of group in each component.
    """

    #logger = logging.getLogger(__name__)

    #find list of unique beads
    beads = []
    for i in range(len(bead_data)):
        for j in range(len(bead_data[i])):
            if bead_data[i][j][0] not in beads:
                beads.append(bead_data[i][j][0])
    beads.sort()

    nui = np.zeros((len(bead_data), len(beads)))
    for i in range(len(bead_data)):
        for j in range(len(bead_data[i])):
            for k in range(np.size(beads)):
                if bead_data[i][j][0] == beads[k]:
                    nui[i, k] = bead_data[i][j][1]
    return beads, nui

######################################################################
#                                                                    #
#                  Parameter Fitting Data                            #
#                                                                    #
######################################################################
def process_param_fit_inputs(thermo_dict):

    """
    Process parameter fitting information.

    Parameters
    ----------
    thermo_dict : dict
        Dictionary of instructions for thermodynamic calculations or parameter fitting. This dictionary is directly from the input file.

        - opt_params (dict) - Parameters used in basin fitting algorithm

            - fit_bead (str) - Name of bead whose parameters are being fit, should be in bead list of beadconfig
            - fit_params (list[str]) - This list of contains the name of the parameter being fit (e.g. epsilon). See EOS documentation for supported parameter names. Cross interaction parameter names should be composed of parameter name and the other bead type, separated by an underscore (e.g. epsilon_CO2).
            - *name* (dict) - Dictionary of a data set that the parameters are fit to. Each dictionary is added to the exp_data dictionary before being passed to the fitting algorithm. Each *name* is used as the key in exp_data. *name* is an arbitrary string used to identify the data set and used later in reporting objective function values during the fitting process. See data type objects for more details.
            - *_bounds (list[float]), Optional - This list contains the minimum and maximum of the parameter from a parameter listed in fit_params, represented in place of the asterisk. See input file instructions for more information.
            - beadparams0 (list[float]), Optional - Initial guess in parameters being fit. Should be the same length at fit_params and contain a reasonable guess for each parameter. If this is not provided, a guess is made based on the type of parameter from eos object.

        - datatype (str) - One of the supported data type objects to fit parameters

    Returns
    -------
    new_thermo_dict : dict
        Dictionary of instructions for thermodynamic calculations or parameter fitting. This dictionary is reformatted and includes imported data. Dictionary values below are altered before being passed on, all other key and value sets are blindly passed on.

        - opt_params (dict) - Parameters used in basin fitting algorithm

            - fit_bead (str) - Name of bead whose parameters are being fit, should be in bead list of beadconfig
            - fit_params (list[str]) - This list of contains the name of the parameter being fit (e.g. epsilon). See EOS documentation for supported parameter names. Cross interaction parameter names should be composed of parameter name and the other bead type, separated by an underscore (e.g. epsilon_CO2).
            - bounds (numpy.ndarray) - List of lists of length two, of length equal to fit_params. If no bounds were given then the default parameter boundaries are [0,1e+4], else bounds given as *_bounds in input file are used.
            - beadparams0 (list[float]), Optional - Initial guess in parameter. If one is not provided, a guess is made based on the type of parameter from eos object.

        - exp_data (dict) - This dictionary is made up of a dictionary for each data set that the parameters are fit to. Each key is an arbitrary string used to identify the data set and used later in reporting objective function values during the fitting process. See data type objects for more details.

            - name (str) - One of the supported data type objects to fit parameters
    """

    logger = logging.getLogger(__name__)

    # Initial new dictionary that will have dictionary for extracted data
    new_thermo_dict = {"exp_data":{}}

    for key, value in thermo_dict.items():
        if key == "opt_params": 
            new_opt_params = {}
            keys_del = []
            new_opt_params["bounds"] = [[0,1e+4] for x in range(len(value["fit_params"]))]
            for key2, value2 in value.items():
                if key2 == "fit_bead":
                    new_opt_params["fit_bead"] = value["fit_bead"]
                elif key2 == "fit_params":
                    new_opt_params["fit_params"] = value["fit_params"]
                elif key2 == "beadparams0":
                    new_opt_params["beadparams0"] = value["beadparams0"]
                elif "bounds" in key2:
                    tmp  = key2.replace("_bounds","")
                    ind = value["fit_params"].index(tmp)
                    new_opt_params["bounds"][ind] = value2
                else:
                    continue
                keys_del.append(key2)
            for key2 in keys_del:
                value.pop(key2,None)

            if value:
               logger.info("The opt_params keys: {}, were not used.".format(", ".join(list(value.keys()))))
            new_thermo_dict[key] = new_opt_params

        elif (type(value) == dict and "datatype" in value):
            new_thermo_dict["exp_data"][key] = process_exp_data(value)

        else:
            new_thermo_dict[key] = value

    # Move opt_params values to thermo_dict
    keys = ["bounds","minimizer_dict"]
    for key in keys:
        if key in new_thermo_dict["opt_params"]:
            new_thermo_dict[key] = new_thermo_dict["opt_params"][key]
            new_thermo_dict["opt_params"].pop(key,None)

    test1 = set(["exp_data","opt_params"]).issubset(list(new_thermo_dict.keys()))
    test2 = set(["fit_bead","fit_params"]).issubset(list(new_thermo_dict["opt_params"].keys()))
    if not all([test1,test2]):
        raise ValueError("An exp_data and opt_params dictionary with, fit_beads and fit_params must be given")

    return new_thermo_dict

######################################################################
#                                                                    #
#                  Process Experimental Data                         #
#                                                                    #
######################################################################
def process_exp_data(exp_data_dict):

    """
    Convert experimental data dictionary into form used by parameter fitting module. 

    Note that there should be one dictionary per data set. All data is extracted from data files.

    Parameters
    ----------
    exp_data : dict
       This dictionary is made up of a dictionary for each data set that the parameters are fit to. Each key is an arbitrary string used to identify the data set and used later in reporting objective function values during the fitting process. See data type objects for more details. Dictionary values below are altered before being passed on, all other key and value sets are blindly passed on.

       - datatype (str) - One of the supported data type objects to fit parameters
       - file (str) - File name in current working directory, or path to desired experimental data. See experimental data page for examples of acceptable format.

    Returns
    -------
    exp_data : dict
        Reformatted dictionary of experimental data. This dictionary is made up of a dictionary for each data set that the parameters are fit to. Each key is an arbitrary string used to identify the data set and used later in reporting objective function values during the fitting process. See data type objects for more details. Dictionary values below are altered from input dictionary, all other key and value sets are blindly passed on, or extracted from data file with process_exp_data_file function.

        - name (str) - One of the supported data type objects to fit parameters
    """

    #logger = logging.getLogger(__name__)

    exp_data = {}
    for key, value in exp_data_dict.items():
        if key == "datatype":
            exp_data["name"] = value
        elif key == "file":
            file_dict = process_exp_data_file(value)
            exp_data.update(file_dict)
        else:
            exp_data[key] = value

    return exp_data

######################################################################
#                                                                    #
#                  Process Experimental Data                         #
#                                                                    #
######################################################################
def process_exp_data_file(fname):

    """
    Import data file and convert columns into dictionary entries.

    The headers in the file are the dictionary keys. The top line is skipped, and column headers are the second line. Note that column headers should be thermo properties (e.g. T, P, x1, x2, y1, y2) without suffixes. Mole fractions x1, x2, ... should be in the same order as in the beadconfig line of the input file. No mole fractions should be left out.

    Parameters
    ----------
    fname : str
        File name or path to experimental data file

    Returns
    -------
    file_dict : dict
        Dictionary of experimental data from file.
    """

    #logger = logging.getLogger(__name__)

    data = np.transpose(np.genfromtxt(fname, delimiter=',',names=True,skip_header=1))
    file_dict = {name:data[name] for name in data.dtype.names}
    
    # Sort through properties
    key_del = []
    xi, yi, zi = [[],[],[]]
    for key, value in file_dict.items():
       if "#" in key:
           key.replace("#","").replace(" ","")

       # Assuming mole fractions are listed starting at x1 and continue in order
       if key.startswith("x"): 
           xi.append(value)
       elif key.startswith("y"):
           yi.append(value)
       elif key.startswith("z"):
           zi.append(value)
       else:
           continue
       key_del.append(key)

    for key in key_del:
        file_dict.pop(key,None)

    if xi: file_dict["xi"] = np.transpose(np.array([np.array(x) for x in xi]))
    if yi: file_dict["yi"] = np.transpose(np.array([np.array(y) for y in yi]))
    if zi: file_dict["zi"] = np.transpose(np.array([np.array(z) for z in zi]))

    return file_dict

######################################################################
#                                                                    #
#                  Write Thermodynamic Output                         #
#                                                                    #
######################################################################
def writeout_thermodict(output_dict,calctype,output_file="thermo_output.txt"):
    """
    Write out result of thermodynamic calculation.

    Import dictionary of both input and output data to produce a file. A line in the top clarifies the calculation type done.

    Parameters
    ----------
    output_dict : dict
        Dictionary of given and calculated information from thermodynamic module
    output_file : str, Optional, default: thermo_output.txt
        Name of output file

    Returns
    -------
    File of data saved to current directory
    """

    # Define units
    units = {"T":"K","P":"Pa","Psat":"Pa","rhol":"mol/m^3","rhov":"mol/m^3"}

    # Make comment line
    comment = "# This data was generated in DESPASITO using the thermodynamic calculation: "+calctype
    
    # Make results matrix
    keys = []
    matrix = []
    for key, value in output_dict.items():
        tmp_matrix = np.transpose(np.array(value))
        if len(tmp_matrix.shape)==1:
            keys.append(key)
            matrix.append(np.array(value))
        else:
            for i in range(len(tmp_matrix)):
                keys.append(key+str(i+1))
                matrix.append(np.array(tmp_matrix[i]))
    matrix = np.transpose(np.array(matrix))

    # Make header line
    header = "#"
    for key in keys:
        if key in units:
            unit = " [{}]".format(units[key])
        else:
            unit = ""
        header += " {}{},".format(key,unit)

    # Write to file
    with open(output_file,"w") as f:
        f.write(comment+"\n")
        f.write(header+"\n")
        for row in matrix:
            f.write((' {},' * len(row)).format(*row)+"\n")
            
######################################################################
#                                                                    #
#                  Write Parameter Fitting Output                    #
#                                                                    #
######################################################################
def writeout_fitdict(output_dict,eos,output_file="fit_output.txt"):
    """
    Write out result of fitting calculation.

    Import fitting results dictionary to write out fitting results. In the future, the EOS dictionary will be used to generate new parameter .json files.

    Parameters
    ----------
    output_dict : dict
        Dictionary of given and calculated information from thermodynamic module.
    eos : obj
        Equation of state output that writes pressure, max density, chemical potential, updates parameters, and evaluates objective functions. For parameter fitting algorithm See equation of state documentation for more details.
    output_file : str, Optional, default: thermo_output.txt
        Name of output file

    Returns
    -------
    File of data saved to current directory
    """

    header = "DESPASITO was used to fit parameters for the bead {} Obj. Value: {}\n".format(output_dict["fit_bead"],output_dict["objective_value"]) + "Parameter, Value\n"
    with open(output_file,"w") as f:
        f.write(header)
        for i in range(len(output_dict["fit_parameters"])):
            f.write("{}, {}".format(output_dict["fit_parameters"][i],output_dict["final_parameters"][i]))

    # Add function to sort through and write out values
    # step 1, Update parameters for EOS object
    # step 2, write out libraries






"""
This module contains our thermodynamic calculations. Calculation of pressure, chemical potential, and max density are handled by an eos object so that these functions can be used with any EOS. The thermo module contains a series of wrapper to handle the inputs and outputs of these functions.
    
"""

import numpy as np
from scipy import interpolate
import scipy.optimize as spo
from scipy.ndimage.filters import gaussian_filter1d
import copy
#import matplotlib.pyplot as plt
import logging

######################################################################
#                                                                    #
#                     Calculate Critical Parameters                  #
#                                                                    #
######################################################################
def calc_CC_Pguess(xilist, Tlist, CriticalProp):
    r"""
    Computes the Mie parameters of a mixture from the mixed critical properties of the pure components. 

    From: Mejia, A., C. Herdes, E. Muller. Ind. Eng. Chem. Res. 2014, 53, 4131-4141
    
    Parameters
    ----------
    xilist : list[list[xi]]
        List of different Mole fractions of each component, sum(xi) should equal 1.0 for each set. Each set of components corresponds to a temperature in Tlist.
    Tlist : list[float]
        Temperature of the system corresponding to composition in xilist [K]
    CriticalProp : list[list]
        List of critical properties :math:`T_C`, :math:`P_C`, :math:`\omega`, :math:`\rho_{0.7}`, :math:`Z_C`, :math:`V_C`, and molecular weight, where each of these properties is a list of values for each bead.
    
    Returns
    -------
    Psatm : list[float]
        A list of guesses in pressure based on critical properties, of the same length as xilist and Tlist [Pa]
    """

    logger = logging.getLogger(__name__)

    Tc, Pc, omega, rho_7, Zc, Vc, M = CriticalProp

    ############## Calculate Mixed System Mie Parameters
    flag = 0
    if all(-0.847 > x > 0.2387 for x in omega):
        flag = 1
        logger.warning("Omega is outside of the range that these correlations are valid")

    a = [14.8359, 22.2019, 7220.9599, 23193.4750, -6207.4663, 1732.9600]
    b = [0.0, -6.9630, 468.7358, -983.6038, 914.3608, -1383.4441]
    c = [0.1284, 1.6772, 0.0, 0.0, 0.0, 0.0]
    d = [0.0, 0.4049, -0.1592, 0.0, 0.0, 0.0]
    j = [1.8966, -6.9808, 10.6330, -9.2041, 4.2503, 0.0]
    k = [0.0, -1.6205, -0.8019, 1.7086, -0.5333, 1.0536]

    Tcm, Pcm, sigma, epsilon, Psatm = [[] for x in range(5)]

    i = 0
    jj = 1

    if flag == 1:
        Psatm = np.nan
    elif flag == 0: 
        for kk, xi in enumerate(xilist):
            # Mixture alpha
            omegaij = xi[i] * omega[i] + xi[jj] * omega[jj]
            tmp1 = np.sum([a[ii] * omegaij**ii for ii in range(6)])
            tmp2 = np.sum([b[ii] * omegaij**ii for ii in range(6)])
            l_r = tmp1 / (1. + tmp2)
            C = (l_r / (l_r - 6.)) * (l_r / 6.)**(6. / (l_r - 6.))
            al_tmp = C * (1. / 3. - 1. / (l_r - 3.))
            # Mixture Critical Properties Stewart-Burkhardt-Voo
            K = xi[i] * Tc[i] / Pc[i]**.5 + xi[jj] * Tc[jj] / Pc[jj]**.5
            tmp1 = xi[i] * Tc[i] / Pc[i] + xi[jj] * Tc[jj] / Pc[jj]
            tmp2 = xi[i] * (Tc[i] / Pc[i])**.5 + xi[jj] * (Tc[jj] / Pc[jj])**.5
            J = tmp1 / 3. + 2. / 3. * tmp2**2.
            Tc_tmp = K**2. / J
            Pc_tmp = (K / J)**2.
            # Mixture Pressure Prausnitz-Gunn
            if (Tlist[kk] / Tc[i] > 1. or Tlist[kk] / Tc[jj] > 1.):
                R = 8.3144598  # [kg*m^2/(s^2*mol*K)] Gas constant
                tmp1 = Zc[i] + Zc[jj]
                tmp2 = xi[i] * M[i] * Vc[i] + xi[jj] * M[jj] * Vc[jj]
                Pc_tmp = R * Tc_tmp * tmp1 / tmp2
            # Mixture Molar Density, Plocker Knapp
            Mij = M[i] * xi[i] + M[jj] * xi[jj]
            rho_tmp = 8. / Mij / ((rho_7[i] * M[i])**(-1. / 3.) + (rho_7[jj] * M[jj])**(-1. / 3.))**3.
            Nav = 6.0221415e+23  # avogadros number
    
            tmp1 = np.sum([c[ii] * al_tmp**ii for ii in range(6)])
            tmp2 = np.sum([d[ii] * al_tmp**ii for ii in range(6)])
            Tc_star = tmp1 / (1. + tmp2)
            eps_tmp = Tc_tmp / Tc_star  # [K], multiply by kB to change to energy units
    
            tmp3 = np.sum([j[ii] * al_tmp**ii for ii in range(6)])
            tmp4 = np.sum([k[ii] * al_tmp**ii for ii in range(6)])
            rho_star = tmp3 / (1. + tmp4)
            sig_tmp = (rho_star / rho_tmp / Nav)**(1. / 3.)
    
            # Calculate Psat
            eos_dict['massi'] = np.array([Mij])
            eos_dict['nui'] = np.array([[1]])
            eos_dict['beads'] = ['bead']
            eos_dict['beadlibrary'] = {
                'bead': {
                    'l_r': l_r,
                    'epsilon': eps_tmp,
                    'Vks': 1.0,
                    'Sk': 1.0,
                    'l_a': 6,
                    'mass': Mij,
                    'sigma': sig_tmp
                }
            }
            eos = eos("saft.gamma_mie",**eos_dict)
    
            if (Tlist[kk] < Tc_tmp):
                Psat_tmp, _, _ = calc_Psat(Tlist[kk], np.array([1.0]), eos)
            else:
                Psat_tmp = np.nan
    
            if np.isnan(Psat_tmp):
                Psatm = np.nan
                break
    
            # Save values
            # Nothing is done with rholsat_tmp and rhogsat_tmp
            Tcm.append(Tc_tmp)
            Pcm.append(Pc_tmp)
            sigma.append(sig_tmp)
            epsilon.append(eps_tmp)
            Psatm.append(Psat_tmp)

    return Psatm

######################################################################
#                                                                    #
#                      Pressure-Density Curve                        #
#                                                                    #
######################################################################
def PvsRho(T, xi, eos, minrhofrac=(1.0 / 500000.0), rhoinc=5.0, vspacemax=1.0E-4, **kwargs):

    r"""
    Give an array of density values, calculates the associated pressure given an eos.
    
    Parameters
    ----------
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    minrhofrac : float, Optional, default: (1.0/500000.0)
        Fraction of the maximum density used to calculate, and is equal to, the minimum density of the density array. The minimum density is the reciprocal of the maximum specific volume used to calculate the roots. Passed from inputs to through the dictionary rhodict.
    rhoinc : float, Optional, default: 5.0
        The increment between density values in the density array. Passed from inputs to through the dictionary rhodict.
    vspacemax : float, Optional, default: 1.0E-4
        Maximum increment between specific volume array values. After conversion from density to specific volume, the increment values are compared to this value. Passed from inputs to through the dictionary rhodict.
    maxpack : float, Optional, default: *see EOS method*
        Maximum packing fraction. Passed from inputs to through the dictionary rhodict.

    Returns
    -------
    vlist : numpy.ndarray
        Specific volume array. Length depends on values in rhodict [:math:`m^3`/mol]
    Plist : numpy.ndarray
        Pressure associated with specific volume of system with given temperature and composition [Pa]
    """

    #logger = logging.getLogger(__name__)

    if type(xi) == list:
        xi = np.array(xi)

    #estimate the maximum density based on the hard sphere packing fraction, part of EOS
    maxrho = eos.density_max(xi, T, **kwargs)

    #min rho is a fraction of max rho, such that minrho << rhogassat
    minrho = maxrho * minrhofrac
    #list of densities for P,rho and P,v
    rholist = np.arange(minrho, maxrho, rhoinc)
    #check rholist to see when the spacing
    vspace = (1.0 / rholist[:-1]) - (1.0 / rholist[1:])
    if np.amax(vspace) > vspacemax:
        vspaceswitch = np.where(vspace > vspacemax)[0][-1]
        rholist_2 = 1.0 / np.arange(1.0 / rholist[vspaceswitch + 1], 1.0 / minrho, vspacemax)[::-1]
        rholist = np.append(rholist_2, rholist[vspaceswitch + 2:])

    #compute Pressures (Plist) for rholist
    Plist = eos.P(rholist, T, xi)

    #Flip Plist and rholist arrays
    Plist = Plist[:][::-1]
    rholist = rholist[:][::-1]
    vlist = 1.0 / rholist

    return vlist, Plist


######################################################################
#                                                                    #
#                      Pressure-Volume Spline                        #
#                                                                    #
######################################################################
def PvsV_spline(vlist, Plist):
    r"""
    Fit arrays of specific volume and pressure values to a cubic Univariate Spline.
    
    Parameters
    ----------
    vlist : numpy.ndarray
        Specific volume array. Length depends on values in rhodict [:math:`m^3`/mol]
    Plist : numpy.ndarray
        Pressure associated with specific volume of system with given temperature and composition [Pa]
    
    Returns
    -------
    Pvspline : obj
        Function object of pressure vs. specific volume
    roots : list
        List of specific volume roots. Subtract a system pressure from the output of Pvsrho to find density of vapor and/or liquid densities.
    extrema : list
        List of specific volume values corresponding to local minima and maxima.
    """

    #logger = logging.getLogger(__name__)

    Psmoothed = gaussian_filter1d(Plist, sigma=.5)

    Pvspline = interpolate.InterpolatedUnivariateSpline(vlist, Psmoothed)
    roots = Pvspline.roots().tolist()
    Pvspline = interpolate.InterpolatedUnivariateSpline(vlist, Psmoothed, k=4)
    extrema = Pvspline.derivative().roots().tolist()
    if extrema: 
        if len(extrema) > 2: extrema = extrema[0:2]

    if len(roots) ==2:
        slope, root2 = np.polyfit(vlist[-4:], Plist[-4:], 1)
        roots = np.append(roots,[root2])

    #PvsV_plot(vlist, Plist, Pvspline, markers=extrema)

    return Pvspline, roots, extrema

######################################################################
#                                                                    #
#                      Pressure-Volume Spline                        #
#                                                                    #
######################################################################
def PvsV_plot(vlist, Plist, Pvspline, markers=[]):
    r"""
    Plot pressure vs. specific volume.
    
    Parameters
    ----------
    vlist : numpy.ndarray
        Specific volume array. Length depends on values in rhodict [:math:`m^3`/mol]
    Plist : numpy.ndarray
        Pressure associated with specific volume of system with given temperature and composition [Pa]
    Pvspline : obj
        Function object of pressure vs. specific volume
    markers : list, Optional, default: []
        List of plot markers used in plot
    """

    #logger = logging.getLogger(__name__)

    plt.figure(1)
    plt.plot(vlist,Plist,label="Orig.")
    plt.plot(vlist,Pvspline(vlist),label="Smoothed")
    plt.plot([vlist[0], vlist[-1]],[0,0],"k")
    for k in range(len(markers)):
        plt.plot([markers[k], markers[k]],[min(Plist),max(Plist)],"k")
    plt.xlabel("Specific Volume [$m^3$/mol]"), plt.ylabel("Pressure [Pa]")
    plt.legend(loc="best")
    plt.tight_layout()
    plt.show()

######################################################################
#                                                                    #
#                              Calc Psat                             #
#                                                                    #
######################################################################
def calc_Psat(T, xi, eos, rhodict={}):
    r"""
    Computes the saturated pressure, gas and liquid densities for a single component system.
    
    Parameters
    ----------
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    Psat : float
        Saturation pressure given system information [Pa]
    rhov : float
        Density of vapor at saturation pressure [mol/:math:`m^3`]
    rhol : float
        Density of liquid at saturation pressure [mol/:math:`m^3`]
    """

    logger = logging.getLogger(__name__)

    if np.count_nonzero(xi) != 1:
        if np.count_nonzero(xi>0.1) != 1:
            raise ValueError("Multiple components have compositions greater than 10%, check code for source")
        else:
            ind = np.where((xi>0.1)==True)[0]
            raise ValueError("Multiple components have compositions greater than 0. Do you mean to obtain the saturation pressure of {} with a mole fraction of {}?".format(eos._beads[ind],xi[ind]))

    vlist, Plist = PvsRho(T, xi, eos, **rhodict)
    Pvspline, roots, extrema = PvsV_spline(vlist, Plist)

    if (not extrema or len(extrema)<2):
        logger.warning('Error: One of the components is above its critical point, add an exception to setPsat')
        Psat = np.nan
        roots = [1.0, 1.0, 1.0]

    else:
        ind_Pmin1 = np.argwhere(np.diff(Plist) > 0)[0][0]
        ind_Pmax1 = np.argmax(Plist[ind_Pmin1:]) + ind_Pmin1

        Pmaxsearch = Plist[ind_Pmax1]

        Pconverged = 10 # If the pressure is negative (under tension), we search from a value just above vacuum 
        Pminsearch = max(Pconverged, np.amin(Plist[ind_Pmin1:ind_Pmax1]))

        #print(Pminsearch,Pmaxsearch)
        #PvsV_plot(vlist, Plist, Pvspline, markers=extrema)

        #search Pressure that gives equal area in maxwell construction
        Psat = spo.minimize_scalar(eq_area,
                               args=(Plist, vlist),
                               bounds=(Pminsearch * 1.0001, Pmaxsearch * .9999),
                               method='bounded')

        #Using computed Psat find the roots in the maxwell construction to give liquid (first root) and vapor (last root) densities
        Psat = Psat.x
        Pvspline, roots, extrema = PvsV_spline(vlist, Plist-Psat)

        if len(roots) ==2:
            slope, root2 = np.polyfit(vlist[-4:], Plist[-4:]-Psat, 1)
            roots = np.append(roots,[root2])

    #Psat,rholsat,rhogsat
    return Psat, 1.0 / roots[0], 1.0 / roots[2]

######################################################################
#                                                                    #
#                              Eq Area                               #
#                                                                    #
######################################################################
def eq_area(shift, Pv, vlist):
    r"""
    Objective function used to calculate the saturation pressure.

    Note: If the curve hasn't decayed to 0 yet, estimate the remaining area as a triangle. This isn't super accurate but we are just using the saturation pressure to get started.
    
    Parameters
    ----------
    shift : float
        Guess in Psat value used to translate the pressure vs. specific volume curve [Pa]
    Pv : numpy.ndarray
        Pressure associated with specific volume of system with given temperature and composition [Pa]
    vlist : numpy.ndarray
        Specific volume array. Length depends on values in rhodict [:math:`m^3`/mol]

    Returns
    -------
    obj_value : float
        Output of objective function, the addition of the positive area between first two roots, and negative area between second and third roots, quantity squared.

    """

    logger = logging.getLogger(__name__)

    Pvspline, roots, extrema = PvsV_spline(vlist, Pv-shift)

    if len(roots) >=3:
        a = Pvspline.integral(roots[0], roots[1])
        b = Pvspline.integral(roots[1], roots[2])
    elif len(roots) == 2:
        a = Pvspline.integral(roots[0], roots[1])
        # If the curve hasn't decayed to 0 yet, estimate the remaining area as a triangle. This isn't super accurate but we are just using the saturation pressure to get started.
        slope, root2 = np.polyfit(vlist[-4:], Pv[-4:]-shift, 1)
        b = Pvspline.integral(roots[1], vlist[-1]) + (Pv[-1]-shift)*(root2-vlist[-1])/2
    else:
        logger.warning("Pressure curve without cubic properties has wrongly been accepted. Try decreasing minrhofrac")
        #PvsV_plot(vlist, Pv-shift, Pvspline, markers=extrema)

    return (a + b)**2

######################################################################
#                                                                    #
#                              Calc Rho V Full                       #
#                                                                    #
######################################################################
def calc_rhov(P, T, xi, eos, rhodict={}):
    r"""
    Computes vapor density under system conditions.
    
    Parameters
    ----------
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    rhov : float
        Density of vapor at system pressure [mol/:math:`m^3`]
    flag : int
        A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means we should assume ideal gas
    """

    logger = logging.getLogger(__name__)

    vlist, Plist = PvsRho(T, xi, eos, **rhodict)
    Plist = Plist-P
    Pvspline, roots, extrema = PvsV_spline(vlist, Plist)

    logger.debug("    Find rhov: P {} Pa, roots {} m^3/mol".format(P,roots))

    flag_NoOpt = False
    l_roots = len(roots)
    if l_roots == 0:
        if Pvspline(1/vlist[-1]) < 0:
            if not len(extrema):
                flag = 2
                logger.debug("    Flag 2: The T and yi, {} {}, combination produces a critical fluid at this pressure".format(T,xi))
            else:
                flag = 1
                logger.debug("    Flag 1: The T and yi, {} {}, combination produces a liquid at this pressure".format(T,xi))
            rho_tmp = 1/vlist[0]
            flag_NoOpt = True
        elif min(Plist)+P > 0:
            rho_tmp = np.nan
            if not len(extrema):
                flag = 4
                logger.debug("    Flag 2: The T and yi, {} {}, combination produces a critical fluid at this pressure".format(T,xi))
            else:
                flag = 4
                logger.debug("    Flag 0: This T and xi, {} {}, combination produces a vapor at this pressure. Warning! approaching critical fluid".format(T,xi))
        else:
            logger.warning("    Flag 3: The T and yi, {} {}, won't produce a fluid (vapor or liquid) at this pressure".format(T,xi))
            flag = 3
            #PvsV_plot(vlist, Plist, Pvspline, markers=extrema)
            rho_tmp = np.nan
    elif l_roots == 1:
        if not len(extrema):
            flag = 2
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 2: The T and yi, {} {}, combination produces a critical fluid at this pressure".format(T,xi))
        elif (Pvspline(roots[0])+P) > (Pvspline(max(extrema))+P):
            #logger.debug("Extrema: {}".format(extrema))
            #logger.debug("Roots: {}".format(roots))
            flag = 1
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 1: The T and yi, {} {}, combination produces a liquid at this pressure".format(T,xi))
        elif len(extrema) > 1:
            flag = 0
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 0: This T and yi, {} {}, combination produces a vapor at this pressure. Warning! approaching critical fluid".format(T,xi))
    elif l_roots == 2:
        if (Pvspline(roots[0])+P) < 0.:
            flag = 1
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 1: This T and xi, {} {}, combination produces a liquid under tension at this pressure".format(T,xi))
        else:
            flag = 4
            rho_tmp = np.nan
            logger.debug("    Flag 4: There should be a third root! Assume ideal gas P: {}".format(P))
    else: # 3 roots
        rho_tmp = 1.0 / roots[2]
        flag = 0

    if flag in [0,2]: # vapor or critical fluid
        tmp = [rho_tmp*.99, rho_tmp*1.01]
        if (Pdiff(tmp[0],P, T, xi, eos)*Pdiff(tmp[1],P, T, xi, eos))<0:
            rho_tmp = spo.brentq(Pdiff, tmp[0], tmp[1], args=(P, T, xi, eos), rtol=0.0000001)
        else:
            if Plist[0] < 0:
                logger.warning("Density value could not be bounded with (rhomin,rhomax), {}. Using approximate density value".format(tmp))
            elif not flag_NoOpt:
                rho_tmp = spo.root(Pdiff, rho_tmp, args=(P, T, xi, eos), method="hybr", tol=0.0000001)
                rho_tmp = rho_tmp.x[0]

    logger.info("    Vapor Density: {} mol/m^3, flag {}".format(rho_tmp,flag))

    # Flag: 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means we should assume ideal gas
    return rho_tmp, flag


######################################################################
#                                                                    #
#                              Calc Rho L Full                       #
#                                                                    #
######################################################################
def calc_rhol(P, T, xi, eos, rhodict={}):
    r"""
    Computes liquid density under system conditions.
    
    Parameters
    ----------
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    rhol : float
        Density of liquid at system pressure [mol/:math:`m^3`]
    flag : int
        A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true
    """

    logger = logging.getLogger(__name__)

    # Get roots and local minima and maxima 
    vlist, Plist = PvsRho(T, xi, eos, **rhodict)
    Plist = Plist-P
    Pvspline, roots, extrema = PvsV_spline(vlist, Plist)

    logger.debug("    Find rhol: P {} Pa, roots {} m^3/mol".format(P,str(roots)))
    flag_NoOpt = False

    if extrema:
        if len(extrema) == 1:
            logger.warning("One extrema at {}, assume weird minima behavior. Check your parameters.".format(1/extrema[0])) 

    # Assess roots, what is the liquid density
    l_roots = len(roots)
    if l_roots == 0: # zero roots
        if Pvspline(1/vlist[-1]):
            if not len(extrema):
                flag = 2
                logger.debug("    Flag 2: The T and yi, {} {}, combination produces a critical fluid at this pressure".format(T,xi))
            else:
                flag = 1
                logger.debug("    Flag 1: The T and yi, {} {}, combination produces a liquid at this pressure".format(T,xi))
            rho_tmp = 1/vlist[1]
            flag_NoOpt = True
        elif min(Plist)+P > 0:
            rho_tmp = np.nan
            if not len(extrema):
                flag = 4
                logger.debug("    Flag 2: The T and yi, {} {}, combination produces a critical fluid at this pressure".format(T,xi))
            else:
                flag = 4
                logger.debug("    Flag 0: This T and xi, {} {}, combination produces a vapor at this pressure. Warning! approaching critical fluid".format(T,xi))
        else:
            flag = 3
            logger.error("    Flag 3: The T and xi, {} {}, won't produce a fluid (vapor or liquid) at this pressure".format(str(T),str(xi)))
            rho_tmp = np.nan
            #PvsV_plot(vlist, Plist+P, Pvspline, markers=extrema)
    elif l_roots == 2: # 2 roots
        if (Pvspline(roots[0])+P) < 0.:
            flag = 1
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 1: This T and xi, {} {}, combination produces a liquid under tension at this pressure".format(T,xi))
        else: # There should be three roots, but the values of specific volume don't go far enough to pick up the last one
            flag = 1
            rho_tmp = 1.0 / roots[0]
    elif l_roots == 1: # 1 root
        if not len(extrema):
            flag = 2
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 2: The T and xi, {} {}, combination produces a critical fluid at this pressure".format(T,xi))
        elif (Pvspline(roots[0])+P) > (Pvspline(max(extrema))+P):
            flag = 1
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 1: The T and xi, {} {}, combination produces a liquid at this pressure".format(T,xi))
        elif len(extrema) > 1:
            flag = 0
            rho_tmp = 1.0 / roots[0]
            logger.debug("    Flag 0: This T and xi, {} {}, combination produces a vapor at this pressure. Warning! approaching critical fluid".format(T,xi))
    else: # 3 roots
        rho_tmp = 1.0 / roots[0]
        flag = 1

    if flag in [1,2]: # liquid or critical fluid
        tmp = [rho_tmp*.99, rho_tmp*1.01]
        if (Pdiff(tmp[0],P, T, xi, eos)*Pdiff(tmp[1],P, T, xi, eos))<0:
            rho_tmp = spo.brentq(Pdiff, tmp[0], tmp[1], args=(P, T, xi, eos), rtol=0.0000001)
        else:
            if Plist[0] < 0:
                logger.warning("Density value could not be bounded with (rhomin,rhomax), {}. Using approximate density value".format(tmp))
            elif not flag_NoOpt:
                rho_tmp = spo.root(Pdiff, rho_tmp, args=(P, T, xi, eos), method="hybr", tol=1e-7)
                rho_tmp = rho_tmp.x[0]
    logger.info("    Liquid Density: {} mol/m^3, flag {}".format(rho_tmp,flag))

    # Flag: 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true
    return rho_tmp, flag

######################################################################
#                                                                    #
#                              Calc Pdiff                            #
#                                                                    #
######################################################################
def Pdiff(rho, Pset, T, xi, eos):
    """
    Calculate difference between set point pressure and computed pressure for a given density.
    
    Parameters
    ----------
    rho : float
        Density of system [mol/:math:`m^3`]
    Pset : float
        Guess in pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    
    Returns
    -------
    Pdiff : float
        Difference in set pressure and predicted pressure given system conditions.
    """

    #logger = logging.getLogger(__name__)

    Pguess = eos.P(rho, T, xi)

    return (Pguess - Pset)

######################################################################
#                                                                    #
#                          Calc phi vapor                            #
#                                                                    #
######################################################################
def calc_phiv(P, T, yi, eos, rhodict={}):
    r"""
    Computes vapor fugacity coefficient under system conditions.
    
    Parameters
    ----------
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    yi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    phiv : float
        Fugacity coefficient of vapor at system pressure
    rhov : float
        Density of vapor at system pressure [mol/:math:`m^3`]
    flag : int
        Flag identifying the fluid type. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means ideal gas is assumed
    """

    logger = logging.getLogger(__name__)

    rhov, flagv = calc_rhov(P, T, yi, eos, rhodict)
    if flagv == 4:
        phiv = np.ones_like(yi)
        rhov = 0.
        logger.info("    rhov set to 0.")
    elif flagv == 3:
        phiv = np.array([np.nan,np.nan])
    else:
        phiv = eos.fugacity_coefficient(P, np.array([rhov]), yi, T)

        #muiv = eos._chemicalpotential_old(P, np.array([rhov]), yi, T)
        #phiv = np.exp(muiv)

    #logger.debug("    Vapor Fugacity Coefficients {}".format(phiv))

    return phiv, rhov, flagv

######################################################################
#                                                                    #
#                         Calc phi liquid                            #
#                                                                    #
######################################################################
def calc_phil(P, T, xi, eos, rhodict={}):
    r"""
    Computes liquid fugacity coefficient under system conditions.
    
    Parameters
    ----------
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    phil : float
        Fugacity coefficient of liquid at system pressure
    rhol : float
        Density of liquid at system pressure [mol/:math:`m^3`]
    flag : int
        Flag identifying the fluid type. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true.
    """

    logger = logging.getLogger(__name__)

    rhol, flagl = calc_rhol(P, T, xi, eos, rhodict)
    if flagl == 4:
        phil = np.ones(len(xi))
        rhol = 0.
        logger.info("    rhol set to 0.")
    elif flagl == 3:
        phil = np.array([np.nan,np.nan])
    else:
        phil = eos.fugacity_coefficient(P, np.array([rhol]), xi, T)

        #muil = eos._chemicalpotential(P, np.array([rhol]), xi, T)
        #phil = np.exp(muil)

    #logger.debug("    Liquid Fugacity Coefficients {}".format(phil))

    return phil, rhol, flagl

######################################################################
#                                                                    #
#                          Calc xi                                   #
#                                                                    #
######################################################################
def calc_yi(xi, phil, phiv):
    r"""
    Computes vapor fugacity coefficient under system conditions.
    
    Parameters
    ----------
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    phil : float
        Fugacity coefficient of liquid at system pressure
    phiv : float
        Fugacity coefficient of vapor at system pressure

    Returns
    -------
    yi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    """

    yi = np.zeros(len(xi))
    ind = np.where(xi != 0.0)[0]
    for i in ind:
        yi[i] = xi[i] * phil[i] / phiv[i]

    return yi

######################################################################
#                                                                    #
#                          Calc yi                                   #
#                                                                    #
######################################################################
def calc_xi(yi, phiv, phil):
    r"""
    Computes vapor fugacity coefficient under system conditions.
    
    Parameters
    ----------
    yi : numpy.ndarray
        Mole fraction of each component, sum(yi) should equal 1.0
    phiv : float
        Fugacity coefficient of vapor at system pressure
    phil : float
        Fugacity coefficient of liquid at system pressure

    Returns
    -------
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    """

    xi = np.zeros(len(yi))
    ind = np.where(yi != 0.0)[0]
    for i in ind:
        xi[i] = yi[i] * phiv[i] / phil[i]

    return xi

######################################################################
#                                                                    #
#                              Calc P range                          #
#                                                                    #
######################################################################
def calc_Prange_xi(T, xi, yi, eos, rhodict={}, Pmin=1000, zi_opts={}):
    r"""
    Obtain min and max pressure values.

    The liquid mole fraction is set and the objective function at each of those values is of opposite sign.
    
    Parameters
    ----------
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    Pmin : float, Optional, default: 1000.0
        Minimum pressure in pressure range that restricts searched space.
    zi_opts : dict, Optional, default: {}
        Options used to solve the inner loop in the solving algorithm

    Returns
    -------
    Prange : list
        List of min and max pressure range
    """

    logger = logging.getLogger(__name__)

    global yi_global

    # Guess a range from Pmin to the local max of the liquid curve
    vlist, Plist = PvsRho(T, xi, eos, **rhodict)
    Pvspline, roots, extrema = PvsV_spline(vlist, Plist)

    #PvsV_plot(vlist, Plist, Pvspline, markers=extrema)

    if not len(extrema):
        Pmax = 100000 # 1 MPa
    else:
        Pmax = max(Pvspline(extrema))
    Parray = [Pmin, Pmax]

    #################### Find Pressure range and Objective Function values

    # Root of min from liquid curve is absolute minimum
    ObjArray = [0, 0]
    yi_range = yi

    maxiter = 200
    for z in range(maxiter):
        # Find Obj Function for Min pressure above
        p = Parray[0]
        phil, rhol, flagl = calc_phil(p, T, xi, eos, rhodict=rhodict)
        if any(np.isnan(phil)):
            logger.error("Estimated minimum pressure is too low.")
            Parray[0] += Pmin
            continue
        yi_range, phiv, flagv = solve_yi_xiT(yi_range, xi, phil, p, T, eos, rhodict=rhodict, **zi_opts)
        ObjArray[0] = (np.sum(xi * phil / phiv) - 1.0)
        logger.debug("Liquid / Vapor Phi: {} /  {}".format(phil,phiv))
        logger.info("Estimated Minimum Pressure: {},  Obj. Func: {}".format(Parray[0],ObjArray[0]))
        if ObjArray[0] > 0:
            break
        else:
            Parray[0] /= 2

    if z == maxiter-1:
        logger.error("Proper minimum pressure for liquid density could not be found")
            

    flag = 0
    for z in range(maxiter):
        if z == 0:
            # Find Obj function for Max Pressure above
            p = Parray[1]
            phil, rhol, flagl = calc_phil(p, T, xi, eos, rhodict=rhodict)
            if any(np.isnan(phil)):
                raise ValueError("Fugacity Coefficient should not be NaN")
            yi_range, phiv, flagv = solve_yi_xiT(yi_range, xi, phil, p, T, eos, rhodict=rhodict, **zi_opts)
            ObjArray[1] = (np.sum(xi * phil / phiv) - 1.0)
            logger.info("Estimate Maximum Pressure: {},  Obj. Func: {}".format(Parray[1],ObjArray[1]))
        else:
            tmp_sum = np.abs(ObjArray[-2] + ObjArray[-1])
            tmp_dif = np.abs(ObjArray[-2] - ObjArray[-1])
            if Parray[-1] == Parray[-2]:
                if flag == 0:
                    p = 2*p
                    flag = 1
                else:
                    raise ValueError("Maximum pressure value could not be found to bound search.")
            elif tmp_dif > tmp_sum:
                if flagv not in [0,2,4]:
                    logger.info("Estimated pressure {}  doesn't produce a vapor, flag={}, Obj Func: {}".format(Parray[-1],flagv,ObjArray[-1]))
                    p = 0.9*Parray[-1]
                else:
                    logger.info("Got the pressure range!")
                    slope = (ObjArray[-1] - ObjArray[-2]) / (Parray[-1] - Parray[-2])
                    intercept = ObjArray[-1] - slope * Parray[-1]
                    Pguess = -intercept / slope
                    break
            else:
                if len(ObjArray) < 2:
                    p = 2 * Parray[-1]
                else:
                    slope = (ObjArray[-1] - ObjArray[-2]) / (Parray[-1] - Parray[-2])
                    intercept = ObjArray[-1] - slope * Parray[-1]
                    p = (-intercept / slope)*1.2 # Add additional 20% to ensure negative value
                    if p < 0.0:
                        ind = ObjArray.index(min(ObjArray))
                        p = Parray[ind]*1.1
            
            phil, rhol, flagl = calc_phil(p, T, xi, eos, rhodict=rhodict)
            if any(np.isnan(phil)):
                raise ValueError("Fugacity coefficient should not be NaN")
            yi_range, phiv, flagv = solve_yi_xiT(yi_range, xi, phil, p, T, eos, rhodict=rhodict, **zi_opts)
            Parray.append(p)
            ObjArray.append(np.sum(xi * phil / phiv) - 1.0)
            logger.info("New Estimate for Maximum Pressure: {},  Obj. Func: {}".format(Parray[-1],ObjArray[-1]))

    if z == maxiter-1:
        logger.error('A change in sign for the objective function could not be found, inspect progress')
        #plt.plot(Parray, ObjArray)
        #plt.plot([Parray[0], Parray[-1]], [0, 0], 'k')
        #plt.ylabel("Obj. Function")
        #plt.xlabel("Pressure / Pa")
        #plt.savefig("CannotFindMax_{}_{}.png".format(T,xi[0]))

    Prange = Parray[-2:]
    ObjRange = ObjArray[-2:]
    logger.info("[Pmin, Pmax]: {}, Obj. Values: {}".format(str(Prange),str(ObjRange)))
    logger.info("Initial guess in pressure: {} Pa".format(Pguess))

    yi_global = yi_range

    return Prange, Pguess

######################################################################
#                                                                    #
#                              Calc P range                          #
#                                                                    #
######################################################################
def calc_Prange_yi(T, xi, yi, eos, rhodict={}, Pmin=1000, zi_opts={}):
    r"""
    Obtain min and max pressure values.

    The vapor mole fraction is set and the objective function at each of those values is of opposite sign.
    
    Parameters
    ----------
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    Pmin : float, Optional, default: 1000.0
        Minimum pressure in pressure range that restricts searched space.
    zi_opts : dict, Optional, default: {}
        Options used to solve the inner loop in the solving algorithm

    Returns
    -------
    Prange : list
        List of min and max pressure range
    """

    logger = logging.getLogger(__name__)

    global xi_global

    # Guess a range from Pmin to the local max of the liquid curve
    vlist, Plist = PvsRho(T, yi, eos, **rhodict)
    Pvspline, roots, extrema = PvsV_spline(vlist, Plist)

    # Calculation the highest pressure possible
    Pmax = max(Pvspline(extrema))
    Parray = [Pmin, Pmax]

    ############################# Test
    pressures = np.linspace(Pmin,Pmax,30)
    pressure2 = np.linspace(pressures[-2],Pmax,20)
    pressures = np.concatenate((pressures,pressure2),axis=0)
    obj_list = []
    for p in pressures:
        phiv, rhov, flagv = calc_phiv(p, T, yi, eos, rhodict=rhodict)
        xi, phil, flagl = solve_xi_yiT(xi, yi, phiv, p, T, eos, rhodict=rhodict, **zi_opts)
        obj_list.append(np.sum(yi * phiv / phil) - 1.0)
    #plt.plot(pressures,obj_list,".-")
    #plt.plot([Pmin,Pmax],[0,0],"k")
    #plt.show()

    #################### Find Pressure range and Objective Function values

    ObjArray = [0, 0]
    xi_range = xi

    for j,i in enumerate([0,1,0]):
        p = Parray[i]
        phiv, rhov, flagv = calc_phiv(p, T, yi, eos, rhodict=rhodict)
        xi_range, phil, flagl = solve_xi_yiT(xi_range, yi, phiv, p, T, eos, rhodict=rhodict, **zi_opts)
        ObjArray[i] = (np.sum(yi * phiv / phil) - 1.0)
        if i == 0:
            logger.info("Estimate Minimum pressure: {},  Obj. Func: {}".format(p,ObjArray[i]))
        elif i == 1:
            if ObjArray[i] < 1e-3:
                ObjArray[i] = 0.0
            logger.info("Estimate Maximum pressure: {},  Obj. Func: {}".format(p,ObjArray[i]))
        # Check pressure range
        if j < 2:
            tmp_sum = np.abs(ObjArray[0] + ObjArray[1])
            tmp_dif = np.abs(ObjArray[0] - ObjArray[1])
            if tmp_dif >= tmp_sum:
                logger.info("Got the pressure range!")
                slope = (ObjArray[1] - ObjArray[-2]) / (Parray[1] - Parray[0])
                intercept = ObjArray[1] - slope * Parray[1]
                Pguess = -intercept / slope
                break
            else:
                newPmin = 10
                if Parray[0] != newPmin:
                    Parray[0] = newPmin
                else:
                    raise ValueError("No VLE data may be found given this temperature and vapor composition. If there are no errors in parameter definitions, consider updating the thermo function 'solve_xi_yiT'.")
            
    Prange = Parray[-2:]
    ObjRange = ObjArray[-2:]
    logger.info("[Pmin, Pmax]: {}, Obj. Values: {}".format(str(Prange),str(ObjRange)))
    logger.info("Initial guess in pressure: {} Pa".format(Pguess))

    xi_global = xi_range

    return Prange, Pguess


######################################################################
#                                                                    #
#                       Solve Yi for xi and T                        #
#                                                                    #
######################################################################
def solve_yi_xiT(yi, xi, phil, P, T, eos, rhodict={}, maxiter=30, tol=1e-6):
    r"""
    Find vapor mole fraction given pressure, liquid mole fraction, and temperature.

    Objective function is the sum of the predicted "mole numbers" predicted by the computed fugacity coefficients. Note that by "mole number" we mean that the prediction will only sum to 1 when the correct pressure is chosen in the outer loop. In this inner loop, we seek to find a mole fraction that is converged to reproduce itself in a prediction. If it hasn't, the new "mole numbers" are normalized into mole fractions and used as the next guess.
    In the case that a guess doesn't produce a gas or critical fluid, we use another function to produce a new guess.
    
    Parameters
    ----------
    yi : numpy.ndarray
        Guess in vapor mole fraction of each component, sum(xi) should equal 1.0
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    phil : float
        Fugacity coefficient of liquid at system pressure
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    maxiter : int, Optional, default: 30
        Maximum number of iteration for both the outer pressure and inner vapor mole fraction loops
    tol : float, Optional, default: 1e-6
        Tolerance in sum of predicted yi "mole numbers"

    Returns
    -------
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(xi) should equal 1.0
    phiv : float
        Fugacity coefficient of vapor at system pressure
    flag : int
        Flag identifying the fluid type. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means ideal gas is assumed
    """

    logger = logging.getLogger(__name__)

    global yi_global

    yi /= np.sum(yi)
    yi_total = [np.sum(yi)]
    flag_check_vapor = True # Make sure we only search for vapor compositions once
    logger.info("    Solve yi: P {}, T {}, xi {}, phil {}".format(P, T, xi, phil))
    for z in range(maxiter):

        yi_tmp = yi/np.sum(yi)

        # Try yi
        phiv, rhov, flagv = calc_phiv(P, T, yi_tmp, eos, rhodict=rhodict)

        if ((any(np.isnan(phiv)) or flagv==1) and flag_check_vapor): # If vapor density doesn't exist
            flag_check_vapor = False
            logger.info("    Composition doesn't produce a vapor, let's find one!")
            if all(yi != 0.):
                yinew = find_new_yi(P, T, phil, xi, eos, rhodict=rhodict)
                phiv, rhov, flagv = calc_phiv(P, T, yinew, eos, rhodict=rhodict)
                yinew = xi * phil / phiv
            else:
                yinew = yi

            if any(np.isnan(phiv)):
                phiv = np.nan
                logger.error("Fugacity coefficient of vapor should not be NaN")
        else:
            yinew = xi * phil / phiv

        ## Check for bouncing between values
        #if len(yi_total) > 3:
        #    tmp1 =  (np.abs(np.sum(yinew)-yi_total[-2]) + np.abs(yi_total[-1]-yi_total[-3]))
        #    tmp2 = (np.abs(np.sum(yinew)-yi_total[-2]) + np.abs(yi_total[-1]-yi_total[-3])) < tol/1000
        #    if (tmp1 < np.abs(np.sum(yinew)-yi_total[-1]) and tmp1 < 1e-5):
        #        # This occurs when the P vs. v curve doesn't cross the 0 axis, there could be a larger problem causing this, but in my experience, it's because the curve is not long enough to converge to zero. Instead of the possible endless increase in vector length and a substantial increase in computational time, we simply set the fugacity coefficient to ideal and the density to 0. When an iteration on our assumption produces a vapor near ideality, it then may predict an ideal gas. This causes the constant back and forth that really isn't that important to solve, as the fugacity coefficients are unity regardless.
        #        logger.info("    yi_total is bouncing between {} and {}, choose the lowest value (outer loop obj. function).".format(np.sum(yinew),yi_total[-1]))
        #        if np.sum(yinew) > yi_total[-1]:
        #            yinew = yi
        #            phiv, rhov, flagv = calc_phiv(P, T, yi_tmp, eos, rhodict=rhodict)

        logger.info("    yi guess {}, yi calc {}, phiv {}".format(yi_tmp,yinew,phiv))
        logger.info("    Old yi_total: {}, New yi_total: {}, Change: {}".format(yi_total[-1],np.sum(yinew),np.sum(yinew)-yi_total[-1])) 

        # Check convergence
        if abs(np.sum(yinew)-yi_total[-1]) < tol:
            ind_tmp = np.where(yi_tmp == min(yi_tmp[yi_tmp>0]))[0] 
            yi2 = yinew/np.sum(yinew)
            if np.abs(yi2[ind_tmp] - yi_tmp[ind_tmp]) / yi_tmp[ind_tmp] < tol:
                yi_global = yi_tmp
                logger.info("    Found yi")
                break

        if z < maxiter-1:
            yi_total.append(np.sum(yinew))
            yi = yinew

    ## If yi wasn't found in defined number of iterations
    yi_tmp = yi/np.sum(yi)
    yinew /= np.sum(yinew)

    ind_tmp = np.where(yi_tmp == min(yi_tmp[yi_tmp>0]))[0]
    if z == maxiter - 1:
        yi2 = yinew/np.sum(yinew)
        tmp = (np.abs(yi2[ind_tmp] - yi_tmp[ind_tmp]) / yi_tmp[ind_tmp])
        logger.warning('    More than {} iterations needed. Error in Smallest Fraction: {} %%'.format(maxiter, tmp*100))
        if tmp > .1: # If difference is greater than 10%
            yinew = find_new_yi(P, T, phil, xi, eos, rhodict=rhodict)
        yinew = spo.least_squares(obj_yi, yinew[0], bounds=(0.,1.), args=(P, T, phil, xi, eos, rhodict))
        yi = yinew.x[0]
        yi = np.array([yi,1-yi])
        obj = obj_yi(yi, P, T, phil, xi, eos, rhodict=rhodict)
        logger.warning('    Find yi with root algorithm, yi {}, obj {}'.format(yi,obj))
    else:
        logger.info("    Inner Loop Final yi: {}, Final Error on Smallest Fraction: {}".format(yi_tmp,np.abs(yi2[ind_tmp] - yi_tmp[ind_tmp]) / yi_tmp[ind_tmp]*100))

    return yi_tmp, phiv, flagv

######################################################################
#                                                                    #
#                       Solve Yi for xi and T                        #
#                                                                    #
######################################################################
def solve_xi_yiT(xi, yi, phiv, P, T, eos, rhodict={}, maxiter=20, tol=1e-6):
    r"""
    Find liquid mole fraction given pressure, vapor mole fraction, and temperature. 

    Objective function is the sum of the predicted "mole numbers" predicted by the computed fugacity coefficients. Note that by "mole number" we mean that the prediction will only sum to 1 when the correct pressure is chosen in the outer loop. In this inner loop, we seek to find a mole fraction that is converged to reproduce itself in a prediction. If it hasn't, the new "mole numbers" are normalized into mole fractions and used as the next guess.
    In the case that a guess doesn't produce a liquid or critical fluid, we use another function to produce a new guess.
    
    Parameters
    ----------
    xi : numpy.ndarray
        Guess in liquid mole fraction of each component, sum(xi) should equal 1.0
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(xi) should equal 1.0
    phiv : float
        Fugacity coefficient of liquid at system pressure
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    maxiter : int, Optional, default: 20
        Maximum number of iteration for both the outer pressure and inner vapor mole fraction loops
    tol : float, Optional, default: 1e-6
        Tolerance in sum of predicted xi "mole numbers"

    Returns
    -------
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    phil : float
        Fugacity coefficient of liquid at system pressure
    flag : int
        Flag identifying the fluid type. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true
    """

    logger = logging.getLogger(__name__)

    global xi_global

    xi /= np.sum(xi)
    xi_total = [np.sum(xi)]
    logger.info("    Solve xi: P {}, T {}, yi {}, phiv {}".format(P, T, yi, phiv))
    for z in range(maxiter):

        xi_tmp = xi/np.sum(xi)

        # Try xi
        phil, rhol, flagl = calc_phil(P, T, xi_tmp, eos, rhodict=rhodict)

        if (any(np.isnan(phil)) or flagl==0): # If liquid density doesn't exist
            logger.info("    Composition doesn't produce a liquid, let's find one!")
            if all(xi != 0.):
                xinew = find_new_xi(P, T, phiv, yi, eos, rhodict=rhodict)
                phil, rhol, flagl = calc_phil(P, T, xinew, eos, rhodict=rhodict)
                xinew = yi * phiv / phil
            else:
                xinew = xi
            
            if any(np.isnan(phil)):
                phil = np.nan
                logger.error("Fugacity coefficient of liquid should not be NaN")
        else:
            xinew = yi * phiv / phil

        logger.info("    xi guess {}, xi calc {}, phil {}".format(xi_tmp,xinew,phil))
        logger.info("    Old xi_total: {}, New xi_total: {}, Change: {}".format(xi_total[-1],np.sum(xinew),np.sum(xinew)-xi_total[-1]))

        # Check convergence
        if abs(np.sum(xinew)-xi_total[-1]) < tol:
            ind_tmp = np.where(xi_tmp == min(xi_tmp[xi_tmp>0]))[0]
            xi2 = xinew/np.sum(xinew)
            if np.abs(xi2[ind_tmp] - xi_tmp[ind_tmp]) / xi_tmp[ind_tmp] < tol:
                xi_global = xi_tmp
                logger.info("    Found xi")
                break

        if z < maxiter-1:
            xi_total.append(np.sum(xinew))
            xi = xinew

    ## If xi wasn't found in defined number of iterations
    xinew /= np.sum(xinew)

    ind_tmp = np.where(xi == min(xi[xi>0]))[0]
    if z == maxiter - 1:
        xi2 = xinew/np.sum(xinew)
        tmp = (np.abs(xi2[ind_tmp] - xi_tmp[ind_tmp]) / xi_tmp[ind_tmp])
        logger.warning('    More than {} iterations needed. Error in Smallest Fraction: {} %%'.format(maxiter, tmp*100))
        if tmp > .1: # If difference is greater than 10%
            xinew = find_new_xi(P, T, phiv, yi, eos, rhodict=rhodict)
        xinew = spo.least_squares(obj_xi, xinew[0], bounds=(0.,1.), args=(P, T, phiv, yi, eos, rhodict))
        xi = xinew.x[0]
        xi = np.array([xi,1-xi])
        obj = obj_xi(xi, P, T, phiv, yi, eos, rhodict=rhodict)
        logger.warning('    Find xi with root algorithm, xi {}, obj {}'.format(xi,obj))
    else:
        logger.info("    Inner Loop Final xi: {}, Final Error on Smallest Fraction: {}".format(xi_tmp,np.abs(xi2[ind_tmp] - xi_tmp[ind_tmp]) / xi_tmp[ind_tmp]*100))

    return xi_tmp, phil, flagl

######################################################################
#                                                                    #
#                       Find new Yi                                  #
#                                                                    #
######################################################################


def find_new_yi(P, T, phil, xi, eos, rhodict={}):
    r"""
    Search vapor mole fraction combinations for a new estimate that produces a vapor density.
    
    Parameters
    ----------
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    phil : float
        Fugacity coefficient of liquid at system pressure
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    """

    logger = logging.getLogger(__name__)

    yi_ext = np.linspace(0.01,.99,30) # Guess for yi
    obj_ext = []
    flag_ext = [[],[]]

    for yi in yi_ext:
        yi = np.array([yi, 1-yi])
        ####
        phiv, rhov, flagv = calc_phiv(P, T, yi, eos, rhodict=rhodict)
        yinew = xi * phil / phiv
        yinew_total_1 = np.sum(yinew)

        yi2 = yinew/yinew_total_1
        phiv2, rhov2, flagv2 = calc_phiv(P, T, yi2, eos, rhodict=rhodict)
        yinew = xi * phil / phiv2
        yinew_total_2 = np.sum(yinew)

        logger.debug("    yi_totals {} {}".format(yinew_total_1,yinew_total_2))

        obj = yinew_total_1 - yinew_total_2

        flag_ext[0].append(flagv)
        flag_ext[1].append(flagv2)
        ######
    #    obj = obj_yi(yi, P, T, phil, xi, eos, rhodict=rhodict)
        obj_ext.append(abs(obj))
        logger.debug("    Obj yi {} total1 - total2 = {}".format(yi,obj))

    #plt.figure(1)
    #plt.plot(yi_ext,obj_ext,".-b")
    #plt.figure(2)
    #plt.plot(yi_ext,flag_ext[0],".-b")
    #plt.plot(yi_ext,flag_ext[1],".-r")
    #plt.show()

    obj_ext = np.array(obj_ext)
    flag_ext = np.array(flag_ext)

    tmp = np.count_nonzero(~np.isnan(obj_ext))
    logger.debug("    Number of valid mole fractions: {}".format(tmp))
    if tmp == 0:
        yi_tmp = np.nan
        obj_tmp = np.nan
    else:
        # Remove any NaN
        obj_tmp  =  obj_ext[~np.isnan(obj_ext)]
        yi_tmp   =   yi_ext[~np.isnan(obj_ext)]
        flag_tmp = flag_ext[1][~np.isnan(obj_ext)]
 
        # Assess vapor values
        # A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means we should assume ideal gas
        ind = [i for i in range(len(flag_tmp)) if flag_tmp[i] not in [1,3]]
        if ind:
            obj_tmp = [obj_tmp[i] for i in ind]
            yi_tmp = [yi_tmp[i] for i in ind]

        # Choose values with lowest objective function
        ind = np.where(np.abs(obj_tmp)==min(np.abs(obj_tmp)))[0][0]
        obj_tmp = obj_tmp[ind]
        yi_tmp = yi_tmp[ind]

    logger.info("    Found new guess in yi: {}, Obj: {}".format(yi_tmp,obj_tmp))
    if type(yi_tmp) not in [list,np.ndarray]:
        yi_tmp = np.array([yi_tmp, 1-yi_tmp])

    return yi_tmp

######################################################################
#                                                                    #
#                       Find new Yi                                  #
#                                                                    #
######################################################################


def obj_yi(yi, P, T, phil, xi, eos, rhodict={}):
    r"""
    Objective function for solving for stable vapor mole fraction.
    
    Parameters
    ----------
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    phil : float
        Fugacity coefficient of liquid at system pressure
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    obj : numpy.ndarray
        Objective function for solving for vapor mole fractions
    """

    logger = logging.getLogger(__name__)

    if type(yi) == float or len(yi) == 1:
        if type(yi) in [list, np.ndarray]:
            yi = np.array([yi[0], 1-yi[0]])
        else:
            yi = np.array([yi, 1-yi])

    phiv, rhov, flagv = calc_phiv(P, T, yi, eos, rhodict=rhodict)
    yinew = xi * phil / phiv
    yinew_total_1 = np.sum(yinew)

    yi2 = yinew/yinew_total_1
    phiv2, rhov2, flagv2 = calc_phiv(P, T, yi2, eos, rhodict=rhodict)
    yinew = xi * phil / phiv2
    yinew_total_2 = np.sum(yinew)

    logger.debug("    yi_totals {} {}".format(yinew_total_1,yinew_total_2))

    obj = np.abs(yinew_total_1 - yinew_total_2)
    
    return obj

######################################################################
#                                                                    #
#                       Find new Xi                                  #
#                                                                    #
######################################################################


def find_new_xi(P, T, phiv, yi, eos, rhodict={}):
    r"""
    Search liquid mole fraction combinations for a new estimate that produces a liquid density.
        
    Parameters
    ----------
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    phiv : float
        Fugacity coefficient of vapor at system pressure
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole
        
    Returns
    -------
    xi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    """
    
    logger = logging.getLogger(__name__)
    
    xi_ext = np.linspace(0.01,.99,30) # Guess for yi
    obj_ext = []
    flag_ext = [[],[]]
    
    for xi in xi_ext:
        xi = np.array([xi, 1-xi])
        ####
        phil, rhol, flagl = calc_phil(P, T, xi, eos, rhodict=rhodict)
        xinew = yi * phiv / phil
        xinew_total_1 = np.sum(xinew)
        
        xi2 = xinew/xinew_total_1
        phil2, rhol2, flagl2 = calc_phil(P, T, xi2, eos, rhodict=rhodict)
        xinew = yi * phiv / phil2
        xinew_total_2 = np.sum(xinew)
        
        logger.debug("    xi_totals {} {}".format(xinew_total_1,xinew_total_2))
        
        obj = xinew_total_1 - xinew_total_2
        
        flag_ext[0].append(flagl)
        flag_ext[1].append(flagl2)
        ######
        #    obj = obj_yi(yi, P, T, phil, xi, eos, rhodict=rhodict)
        obj_ext.append(abs(obj))
        logger.debug("    Obj xi {} total1 - total2 = {}".format(xi,obj))

    obj_ext = np.array(obj_ext)
    flag_ext = np.array(flag_ext)

    tmp = np.count_nonzero(~np.isnan(obj_ext))
    logger.debug("    Number of valid mole fractions: {}".format(tmp))
    if tmp == 0:
        xi_tmp = np.nan
        obj_tmp = np.nan
    else:
        # Remove any NaN
        obj_tmp  =  obj_ext[~np.isnan(obj_ext)]
        xi_tmp   =   xi_ext[~np.isnan(obj_ext)]
        flag_tmp = flag_ext[1][~np.isnan(obj_ext)]
    
        # Assess vapor values
        # A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means we should assume ideal gas
        ind = [i for i in range(len(flag_tmp)) if flag_tmp[i] not in [0,3,4]]
        if ind:
            obj_tmp = [obj_tmp[i] for i in ind]
            xi_tmp = [xi_tmp[i] for i in ind]
    
        # Choose values with lowest objective function
        ind = np.where(np.abs(obj_tmp)==min(np.abs(obj_tmp)))[0][0]
        obj_tmp = obj_tmp[ind]
        xi_tmp = xi_tmp[ind]
    
    logger.info("    Found new guess in xi: {}, Obj: {}".format(xi_tmp,obj_tmp))

    if type(xi_tmp) not in [list,np.ndarray]:
        xi_tmp = np.array([xi_tmp, 1-xi_tmp])
    
    return xi_tmp

######################################################################
#                                                                    #
#                       Find new Xi                                  #
#                                                                    #
######################################################################


def obj_xi(xi, P, T, phiv, yi, eos, rhodict={}):
    r"""
    Objective function for solving for stable vapor mole fraction.
        
    Parameters
    ----------
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    P : float
        Pressure of the system [Pa]
    T : float
        Temperature of the system [K]
    phiv : float
        Fugacity coefficient of vapor at system pressure
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole
        
    Returns
    -------
    obj : numpy.ndarray
        Objective function for solving for liquid mole fractions
    """
    
    logger = logging.getLogger(__name__)
    
    if type(xi) == float or len(xi) == 1:
        if type(xi) in [list, np.ndarray]:
            xi = np.array([xi[0], 1-xi[0]])
        else:
            xi = np.array([xi, 1-xi])

    phil, rhol, flagl = calc_phil(P, T, xi, eos, rhodict=rhodict)
    xinew = yi * phiv / phil
    xinew_total_1 = np.sum(xinew)
    
    xi2 = xinew/xinew_total_1
    phil2, rhol2, flagl2 = calc_phil(P, T, xi2, eos, rhodict=rhodict)
    xinew = yi * phiv / phil2
    xinew_total_2 = np.sum(xinew)
    
    logger.debug("    xi_totals {} {}".format(xinew_total_1,xinew_total_2))
    
    obj = np.abs(xinew_total_1 - xinew_total_2)
    
    return obj

######################################################################
#                                                                    #
#                              Solve P xT                            #
#                                                                    #
######################################################################
def solve_P_xiT(P, xi, T, eos, rhodict, zi_opts={}):
    r"""
    Objective function used to search pressure values and solve outer loop of P bubble point calculations.
    
    Parameters
    ----------
    P : float
        Guess in pressure of the system [Pa]
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    zi_opts : dict, Optional, default: {}
        Options used to solve the inner loop in the solving algorithm
    

    Returns
    -------
    obj_value : float
        :math:`\sum\frac{x_{i}\{phi_l}{\phi_v}-1`
    """

    logger = logging.getLogger(__name__)

    global yi_global

    if P < 0:
        return 10.0

    logger.info("P Guess: {} Pa".format(P))

    #find liquid density
    phil, rhol, flagl = calc_phil(P, T, xi, eos, rhodict={})

    yinew, phiv, flagv = solve_yi_xiT(yi_global, xi, phil, P, T, eos, rhodict=rhodict, **zi_opts)
    yi_global = yi_global / np.sum(yi_global)

    #given final yi recompute
    phiv, rhov, flagv = calc_phiv(P, T, yi_global, eos, rhodict={})

    Pv_test = eos.P(rhov, T, yi_global)
    obj_value = float((np.sum(xi * phil / phiv) - 1.0))
    logger.info('Obj Func: {}, Pset: {}, Pcalc: {}'.format(obj_value, P, Pv_test[0]))

    return obj_value

######################################################################
#                                                                    #
#                              Solve P yT                            #
#                                                                    #
######################################################################
def solve_P_yiT(P, yi, T, eos, rhodict, zi_opts={}):
    r"""
    Objective function used to search pressure values and solve outer loop of P dew point calculations.
    
    Parameters
    ----------
    P : float
        Guess in pressure of the system [Pa]
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    zi_opts : dict, Optional, default: {}
        Options used to solve the inner loop in the solving algorithm

    Returns
    -------
    obj_value : list
        :math:`\sum\frac{y_{i}\{phi_v}{\phi_l}-1`
    """

    logger = logging.getLogger(__name__)

    global xi_global

    if P < 0:
        return 10.0

    #find liquid density
    phiv, rhov, flagv = calc_phiv(P, T, yi, eos, rhodict={})

    xi_global, phil, flagl = solve_xi_yiT(xi_global, yi, phiv, P, T, eos, rhodict=rhodict, **zi_opts)
    xi_global = xi_global / np.sum(xi_global)

    #given final yi recompute
    phil, rhol, flagl = calc_phil(P, T, xi_global, eos, rhodict={})

    Pv_test = eos.P(rhov, T, xi_global)
    obj_value = (np.sum(xi_global * phil / phiv) - 1.0)
    logger.info('Obj Func: {}, Pset: {}, Pcalc: {}'.format(obj_value, P, Pv_test[0]))

    return obj_value

######################################################################
#                                                                    #
#                   Set Psat for Critical Components                 #
#                                                                    #
######################################################################
def setPsat(ind, eos):
    r"""
    Generate dummy value for component saturation pressure if it is above its critical point.
    
    Parameters
    ----------
    ind : int
        Index of bead that is above critical point
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.

    Returns
    -------
    Psat : float
        Dummy value of saturation pressure [Pa]
    NaNbead : str
        Bead name of the component that is above it's critical point
    """

    logger = logging.getLogger(__name__)

    for j in range(np.size(eos._nui[ind])):
        if eos._nui[ind][j] > 0.0 and eos._beads[j] == "CO2":
            Psat = 10377000.0
        elif eos._nui[ind][j] > 0.0 and eos._beads[j] == "N2":
            Psat = 7377000.0
        elif eos._nui[ind][j] > 0.0 and ("CH4" in eos._beads[j]):
            Psat = 6377000.0
        elif eos._nui[ind][j] > 0.0 and ("CH3CH3" in eos._beads[j]):
            Psat = 7377000.0
        elif eos._nui[ind][j] > 0.0:
            #Psat = np.nan
            Psat = 7377000.0
            NaNbead = eos._beads[j]
            logger.warning("Bead, {}, is above its critical point. Psat is assumed to be {}. To add an exception go to thermodynamics.calc.setPsat".format(NaNbead,Psat))

    if "NaNbead" not in list(locals().keys()):
       NaNbead = "No NaNbead"
       logger.info("No beads above their critical point")

    return Psat, NaNbead 

######################################################################
#                                                                    #
#                              Calc yT phase                         #
#                                                                    #
######################################################################
def calc_yT_phase(yi, T, eos, rhodict={}, zi_opts={}, Pguess=-1, meth="hybr", pressure_opts={}):
    r"""
    Calculate dew point mole fraction and pressure given system vapor mole fraction and temperature.
    
    Parameters
    ----------
    yi : numpy.ndarray
        Vapor mole fraction of each component, sum(yi) should equal 1.0
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    zi_opts : dict, Optional, default: {}
        Options used to solve the inner loop in the solving algorithm
    Pguess : float, Optional, default: -1
        Guess the system pressure at the dew point. A negative value will force an estimation based on the saturation pressure of each component.
    meth : str, Optional, default: "broyden1"
        Choose the method used to solve the dew point calculation
    pressure_opts : dict, Optional, default: {}
        Options used in the given method, "meth", to solve the outer loop in the solving algorithm

    Returns
    -------
    P : float
        Pressure of the system [Pa]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    flagl : int
        Flag identifying the fluid type for the liquid mole fractions, expected is liquid, 1. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true
    flagv : int
        Flag identifying the fluid type for the vapor mole fractions, expected is vapor or 0. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means ideal gas is assumed
    """

    logger = logging.getLogger(__name__)

    global xi_global

    # Estimate pure component vapor pressures
    Psat = np.zeros_like(yi)
    for i in range(np.size(yi)):
        yi_tmp = np.zeros_like(yi)
        yi_tmp[i] = 1.0
        Psat[i], _, _ = calc_Psat(T, yi_tmp, eos, rhodict)
        if np.isnan(Psat[i]):
            Psat[i], NaNbead = setPsat(i, eos)
            if np.isnan(Psat[i]):
                raise ValueError("Component, {}, is beyond it's critical point at {} K. Add an exception to setPsat".format(NaNbead,T))

    # Estimate initial pressure
    if Pguess < 0:
        P=1.0/np.sum(yi/Psat)
    else:
        P = Pguess

    # Estimate initial xi
    if ("xi_global" not in globals() or any(np.isnan(xi_global))):
        xi_global = P * (yi / Psat)
        xi_global /= np.sum(xi_global)
        xi_global = copy.deepcopy(xi_global)
    xi = xi_global 

    #Prange, Pguess = calc_Prange_yi(T, xi, yi, eos, rhodict, zi_opts=zi_opts)
    #logger.info("Given Pguess: {}, Suggested: {}".format(P, Pguess))
    #P = Pguess

    #################### Root Finding without Boundaries ###################
    if meth in ['broyden1', 'broyden2']:
        outer_dict = {'fatol': 1e-5, 'maxiter': 25, 'jac_options': {'reduction_method': 'simple'}}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.root(solve_P_yiT, P, args=(yi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)
    elif meth in ['hybr_broyden1', 'hybr_broyden2']:
        outer_dict = {'fatol': 1e-5, 'maxiter': 25, 'jac_options': {'reduction_method': 'simple'}}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        Pfinal = spo.root(solve_P_yiT, P, args=(yi, T, eos, rhodict, zi_opts), method="hybr")
        Pfinal = spo.root(solve_P_yiT, Pfinal.x, args=(yi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)
    elif meth == 'anderson':
        outer_dict = {'fatol': 1e-5, 'maxiter': 25}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.root(solve_P_yiT, P, args=(yi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)
    elif meth in ['hybr', 'lm', 'linearmixing', 'diagbroyden', 'excitingmixing', 'krylov', 'df-sane']:
        outer_dict = {'xtol': 1e-10}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.root(solve_P_yiT, P, args=(yi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)

#################### Minimization Methods with Boundaries ###################
    elif meth in ["TNC", "L-BFGS-B", "SLSQP"]:
        outer_dict = {}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        if len(Prange) == 2:
            Pfinal = spo.minimize(solve_P_yiT, P, args=(yi, T, eos, rhodict, zi_opts), method=meth, bounds=[tuple(Prange)], options=outer_dict)
        else:
            Pfinal = spo.minimize(solve_P_yiT, P, args=(yi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)

#################### Root Finding with Boundaries ###################
    elif meth == "brent":
        outer_dict = {"rtol":1e-7}
        for key, value in pressure_opts.items():
            if key in ["xtol","rtol","maxiter","full_output","disp"]:
                outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        P = spo.brentq(solve_P_yiT, Prange[0], Prange[1], args=(yi, T, eos, rhodict, zi_opts), **outer_dict)
    elif meth == "least_squares":
        outer_dict = {}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.least_squares(solve_P_xiT, P, bounds=(Prange[0],Prange[1]), args=(xi, T, eos, rhodict, zi_opts), **outer_dict)

    #Given final P estimate
    if meth != "brent":
        P = Pfinal.x
        logger.info("Optimization terminated successfully: {} {}".format(Pfinal.success,Pfinal.message))

    #find vapor density and fugacity
    phiv, rhov, flagv = calc_phiv(P, T, yi, eos, rhodict={})
    phil, rhol, flagl = calc_phil(P, T, xi, eos, rhodict={})
    if "tol" in zi_opts:
        if zi_opts["tol"] > 1e-10:
            zi_opts["tol"] = 1e-10

    obj = solve_P_yiT(P, yi, T, eos, rhodict=rhodict, zi_opts=zi_opts)

    logger.info("Final Output: Obj {}, P {} Pa, flagl {}, xi {}".format(obj,P,flagl,xi_global))

    return P, xi, flagl, flagv, obj

######################################################################
#                                                                    #
#                              Calc xT phase                         #
#                                                                    #
######################################################################
def calc_xT_phase(xi, T, eos, rhodict={}, zi_opts={}, Pguess=-1, meth="hybr", pressure_opts={}):
    r"""
    Calculate bubble point mole fraction and pressure given system liquid mole fraction and temperature.
    
    Parameters
    ----------
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 
    zi_opts : dict, Optional, default: {}
        Options used to solve the inner loop in the solving algorithm
    Pguess : float, Optional, default: -1
        Guess the system pressure at the dew point. A negative value will force an estimation based on the saturation pressure of each component.
    meth : str, Optional, default: "broyden1"
        Choose the method used to solve the dew point calculation
    pressure_opts : dict, Optional, default: {}
        Options used in the given method, "meth", to solve the outer loop in the solving algorithm

    Returns
    -------
    P : float
        Pressure of the system [Pa]
    yi : numpy.ndarray
        Mole fraction of each component, sum(yi) should equal 1.0
    flagv : int
        Flag identifying the fluid type for the vapor mole fractions, expected is vapor or 0. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true, 4 means ideal gas is assumed
    flagl : int
        Flag identifying the fluid type for the liquid mole fractions, expected is liquid, 1. A value of 0 is vapor, 1 is liquid, 2 mean a critical fluid, 3 means that neither is true
    """

    logger = logging.getLogger(__name__)

    global yi_global

    Psat = np.zeros_like(xi)
    for i in range(np.size(xi)):
        xi_tmp = np.zeros_like(xi)
        xi_tmp[i] = 1.0
        Psat[i], _, _ = calc_Psat(T, xi_tmp, eos, rhodict)
        if np.isnan(Psat[i]):
            Psat[i], NaNbead = setPsat(i, eos)
            if np.isnan(Psat[i]):
                logger.error("Component, {}, is beyond it's critical point. Add an exception to setPsat".format(NaNbead))

    # Estimate initial pressure
    if Pguess < 0:
        P=1.0/np.sum(xi/Psat)
    else:
        P = Pguess

    if ("yi_global" not in globals() or any(np.isnan(yi_global))):
        yi_global = xi * Psat / P
        yi_global /= np.sum(yi_global)
        yi_global = copy.deepcopy(yi_global)
        logger.info("Guess yi in calc_xT_phase with Psat: {}".format(yi_global))
    yi = yi_global

#    logger.info("Initial: P: {}, yi: {}".format(Pguess,str(yi)))
#    Pguess, yi = bubblepoint_guess(Pguess, yi, xi, T, phil, eos, rhodict)
#    logger.info("Updated: P: {}, yi: {}".format(Pguess,str(yi)))

    Prange, Pguess = calc_Prange_xi(T, xi, yi, eos, rhodict, zi_opts=zi_opts)
    logger.info("Given Pguess: {}, Suggested: {}".format(P, Pguess))
    P = Pguess

    if meth not in ["brent", "least_squares", "TNC", "L-BFGS-B", "SLSQP", 'hybr', 'lm', 'linearmixing', 'diagbroyden', 'excitingmixing', 'krylov', 'df-sane', 'anderson', 'hybr_broyden1', 'hybr_broyden2', 'broyden1', 'broyden2']:
        logger.error("Optimization method, {}, not supported.".format(meth))

    #################### Root Finding without Boundaries ###################
    if meth in ['broyden1', 'broyden2']:
        outer_dict = {'fatol': 1e-5, 'maxiter': 25, 'jac_options': {'reduction_method': 'simple'}}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.root(solve_P_xiT, P, args=(xi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)
    elif meth in ['hybr_broyden1', 'hybr_broyden2']:
        outer_dict = {'fatol': 1e-5, 'maxiter': 25, 'jac_options': {'reduction_method': 'simple'}}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        Pfinal = spo.root(solve_P_xiT, P, args=(xi, T, eos, rhodict, zi_opts), method="hybr")
        Pfinal = spo.root(solve_P_xiT, Pfinal.x, args=(xi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)
    elif meth == 'anderson':
        outer_dict = {'fatol': 1e-5, 'maxiter': 25}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.root(solve_P_xiT, P, args=(xi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)
    elif meth in ['hybr', 'lm', 'linearmixing', 'diagbroyden', 'excitingmixing', 'krylov', 'df-sane']:
        outer_dict = {}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.root(solve_P_xiT, P, args=(xi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)

#################### Minimization Methods with Boundaries ###################
    elif meth in ["TNC", "L-BFGS-B", "SLSQP"]:
        outer_dict = {}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        if len(Prange) == 2:
            Pfinal = spo.minimize(solve_P_xiT, P, args=(xi, T, eos, rhodict, zi_opts), method=meth, bounds=[tuple(Prange)], options=outer_dict)
        else:
            Pfinal = spo.minimize(solve_P_xiT, P, args=(xi, T, eos, rhodict, zi_opts), method=meth, options=outer_dict)

#################### Root Finding with Boundaries ###################
    elif meth == "brent":
        outer_dict = {"rtol":1e-7}
        for key, value in pressure_opts.items():
            if key in ["xtol","rtol","maxiter","full_output","disp"]:
                outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        P = spo.brentq(solve_P_xiT, Prange[0], Prange[1], args=(xi, T, eos, rhodict, zi_opts), **outer_dict)
    elif meth == "least_squares":
        outer_dict = {}
        for key, value in pressure_opts.items():
            outer_dict[key] = value
        logger.debug("Using the method, {}, with the following options:\n{}".format(meth,outer_dict))
        Pfinal = spo.least_squares(solve_P_xiT, P, bounds=(Prange[0],Prange[1]), args=(xi, T, eos, rhodict, zi_opts), **outer_dict)

    #Given final P estimate
    if meth != "brent":
        P = Pfinal.x
        logger.info("Optimization terminated successfully: {} {}".format(Pfinal.success,Pfinal.message))

    #find liquid density and fugacity
    phil, rhol, flagl = calc_phil(P, T, xi, eos, rhodict={})
    phiv, rhov, flagv = calc_phiv(P, T, yi, eos, rhodict={})
    if "tol" in zi_opts:
        if zi_opts["tol"] > 1e-10:
            zi_opts["tol"] = 1e-10

    obj = solve_P_xiT(P, xi, T, eos, rhodict=rhodict, zi_opts=zi_opts)

    # NoteHere

    logger.info("Final Output: Obj {}, P {} Pa, flagv {}, yi {}".format(obj,P,flagv,yi_global))

    return P, yi_global, flagv, flagl, obj

######################################################################
#                                                                    #
#                              Calc PT phase                         #
#                                                                    #
######################################################################
def calc_PT_phase(xi, T, eos, rhodict={}):
    r"""
    *Not Complete*
    Calculate the PT phase diagram given liquid mole fraction and temperature
    
    Parameters
    ----------
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    T : float
        Temperature of the system [K]
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    P : float
        Pressure of the system [Pa]
    yi : numpy.ndarray
        Mole fraction of each component, sum(yi) should equal 1.0
    """

    logger = logging.getLogger(__name__)

    logger.error("The function, calc_PT_phase, is not yet available.")

  #  Psat = np.zeros_like(xi)
  #  for i in range(np.size(xi)):
  #      xi_tmp = np.zeros_like(xi)
  #      xi_tmp[i] = 1.0
  #      Psat[i], rholsat, rhogsat = calc_Psat(T, xi_tmp, eos, rhodict)
  #      if np.isnan(Psat[i]):
  #          Psat[i], NaNbead = setPsat(i, eos)
  #          if np.isnan(Psat[i]):
  #              logger.error("Component, {}, is beyond it's critical point. Add an exception to setPsat".format(NaNbead))

  #  zi = np.array([0.5, 0.5])

  #  #estimate ki
  #  ki = Psat / P

  #  #estimate beta (not thermodynamic) vapor frac
  #  beta = (1.0 - np.sum(ki * zi)) / np.prod(ki - 1.0)


######################################################################
#                                                                    #
#                              Calc dadT                             #
#                                                                    #
######################################################################
def calc_dadT(rho, T, xi, eos, rhodict={}):
    r"""
    Calculate the derivative of the Helmholtz energy with respect to temperature, :math:`\frac{dA}{dT}`.
    
    Parameters
    ----------
    rho : numpy.ndarray
        Density array. Length depends on values in rhodict [mol/:math:`m^3`]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Liquid mole fraction of each component, sum(xi) should equal 1.0
    eos : obj
        An instance of the defined EOS class to be used in thermodynamic computations.
    rhodict : dict, Optional, default: {}
        Dictionary of options used in calculating pressure vs. mole 

    Returns
    -------
    dadT : numpy.ndarray
        Array of derivative values of Helmholtz energy with respect to temperature
    """

    #logger = logging.getLogger(__name__)

    step = np.sqrt(np.finfo(float).eps) * T * 1000.0

    #computer rho+step and rho-step for better a bit better performance
    Ap = calchelmholtz.calc_A(np.array([rho]), xi, T + step, eos)
    Am = calchelmholtz.calc_A(np.array([rho]), xi, T - step, eos)

    return (Ap - Am) / (2.0 * step)




"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.input_output.readwrite_input as rw
import despasito.fit_parameters as fit
import despasito.fit_parameters.fit_funcs as funcs
import despasito.equations_of_state
import pytest
import sys
import numpy as np

Tlist = [323.2]
xilist = [[0.01,0.99]]
yilist = [[0.97702, 0.02298]]
Plist = [4492927.45]
rho = np.array([19986.78358869])
P = np.array([4099056.163132072])

## EOS Object
beads = ['CH3OH']
nui = np.array([[1.]])
beadlibrary = {'CH3OH': {'epsilon': 375.01, 'l_a': 6.0, 'l_r': 16.352, 'sigma': 3.463e-10, 'Sk': 1.0, 'Vks': 2, 'mass': 0.0310337}}
eos = despasito.equations_of_state.eos(eos="saft.gamma_mie",beads=beads,nui=nui,beadlibrary=beadlibrary)

## Exp Data dict
exp_data = {'Wiley': {'name': 'sat_props', 'calctype': 'sat_props', 'T': np.array([200.]), 'Psat': np.array([6.1000e+01]), 'rhol': np.array([27474.40699]), 'rhov': np.array([3.12109900e-03])}, 'Gibbard': {'name': 'sat_props', 'calctype': 'sat_props', 'T': np.array([288.1506]), 'Psat': np.array([ 9884.4])}}

## Optimization options
opt_params = {"fit_bead" : "CH3OH", "fit_params": ["epsilon"], "epsilon_bounds" : [150.0, 400.0]}

thermo_dict = {"opt_params": opt_params, "exp_data": exp_data, "basin_dict": {"niter": 1, "niter_success": 1}, "beadparams0": [384.0], "minimizer_dict": {"tol": 1e-1, "maxiter": 25}}


def test_fit_import():
#    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.fit_parameters" in sys.modules

def test_fit_1comp(eos=eos,thermo_dict=thermo_dict):

    thermo_dict = rw.process_param_fit_inputs(thermo_dict)
    output = fit.fit(eos,thermo_dict)
        
    assert output["final_parameters"][0]==pytest.approx(384.93,abs=1e-1) and output["objective_value"]<1.0




# -- coding: utf8 --

r"""
    
    EOS object for SAFT-:math:`\gamma`-Mie
    
    Equations referenced in this code are from V. Papaioannou et al J. Chem. Phys. 140 054107 2014
    
"""

import numpy as np
import logging

from . import constants
from . import gamma_mie_funcs as funcs
# Later this line will be in an abstract class file in this directory, and all versions of SAFT will reference it
from despasito.equations_of_state.interface import EOStemplate

# ________________ Saft Family ______________
# NoteHere: Insert SAFT family abstract class in this directory to clean up


class saft_gamma_mie(EOStemplate):

    r"""
    Initialize EOS object for SAFT-:math:`\gamma`-Mie.
    
    All input and calculated parameters are defined as hidden attributes.
    
    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    nui : numpy.ndarray
        :math:`\\nu_{i,k}/k_B`. Array of number of components by number of bead types. Defines the number of each type of group in each component.
    beadlibrary : dict
        A dictionary where bead names are the keys to access EOS self interaction parameters:

        - epsilon: :math:`\epsilon_{k,k}/k_B`, Energy well depth scaled by Boltzmann constant
        - sigma: :math:`\sigma_{k,k}`, Size parameter [m]
        - mass: Bead mass [kg/mol]
        - l_r: :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups of type k
        - l_a: :math:`\lambda^{a}_{k,k}`, Exponent of attractive term between groups of type k\
        - Vks: :math:`V_{k,s}`, Number of groups, k, in component
        - Sk: Optional, :math:`S_{k}`, Shape parameter of group k
        - epsilon*: Optional, Interaction energy between each bead and association site. Asterisk represents string from sitenames.
        - K**: Optional, Bonding volume between each association site. Asterisk represents two strings from sitenames.
        - Nk*: Optional, The number of sites of from list sitenames. Asterisk represents string from sitenames.

    crosslibrary : dict, Optional, default: {}
        Optional library of bead cross interaction parameters. As many or as few of the desired parameters may be defined for whichever group combinations are desired.

        - epsilon: :math:`\epsilon_{k,l}/k_B`, Energy parameter scaled by Boltzmann Constant
        - l_r: :math:`\lambda^{r}_{k,l}`, Exponent of repulsive term between groups of type k and l
        - epsilon*: Optional, Interaction energy between each bead and association site. Asterisk represents string from sitenames.
        - K**: Optional, Bonding volume between each association site. Asterisk represents two strings from sitenames.

    sitenames : list[str], Optional, default: ["H", "e1", "e2"]
        List of unique association sites used among components
        
    Attributes
    ----------
    T : float, default: numpy.nan
        Temperature value is initially defined as NaN for a placeholder until temperature dependent attributes are initialized by using a method of this class.
    
    """

    def __init__(self, kwargs):

        #logger = logging.getLogger(__name__)

        # Self interaction parameters
        self._nui = kwargs['nui']
        self._beads = kwargs['beads']
        self._beadlibrary = kwargs['beadlibrary']

        massi = np.zeros(len(self._nui))
        for i in range(len(self._nui)):
            for k in range(np.size(self._beads)):
                massi[i] += self._nui[i, k] * self._beadlibrary[self._beads[k]]["mass"]
        self._massi = massi

        # Cross interaction parameters
        if 'crosslibrary' in list(kwargs.keys()):
            crosslibrary = kwargs['crosslibrary']
        else:
            crosslibrary = {}

        epsilonkl, sigmakl, l_akl, l_rkl, Ckl = funcs.calc_interaction_matrices(self._beads, self._beadlibrary, crosslibrary=crosslibrary)

        self._crosslibrary = crosslibrary
        self._epsilonkl = epsilonkl
        self._sigmakl = sigmakl
        self._l_akl = l_akl
        self._l_rkl = l_rkl
        self._Ckl = Ckl

        # Association sites
        if 'sitenames' in list(kwargs.keys()):
            self._sitenames = kwargs['sitenames']
        else:
            self._sitenames = ["H", "e1", "e2"]

        epsilonHB, Kklab, nk = funcs.calc_assoc_matrices(self._beads, self._beadlibrary, sitenames=self._sitenames, crosslibrary=self._crosslibrary)

        self._epsilonHB = epsilonHB
        self._Kklab = Kklab
        self._nk = nk

        # Initialize temperature attribute
        self.T = np.nan

    def _temp_dependent_variables(self, T):

        """
        Temperature dependent variables are initialized or updated.
    
        Parameters
        ----------
        T : float, default: numpy.nan
            Temperature of the system [K]
    
        Attributes
        ----------
        T : float, default: numpy.nan
            Temperature of the system
        """

        #logger = logging.getLogger(__name__)

        dkk, dkl, x0kl = funcs.calc_hard_sphere_matricies(self._beads, self._beadlibrary, self._sigmakl, T)
        self.T = T
        self._dkk = dkk
        self._dkl = dkl
        self._x0kl = x0kl

    def _xi_dependent_variables(self, xi):

        """
        Variables dependent on composition are initialized or updated.
    
        Parameters
        ----------
        xi : list[float]
            Mole fraction of component

        """

        #logger = logging.getLogger(__name__)

        Cmol2seg, xsk, xskl = funcs.calc_composition_dependent_variables(xi, self._nui, self._beads, self._beadlibrary)
        self._Cmol2seg = Cmol2seg
        self._xsk = xsk
        self._xskl = xskl

    def P(self, rho, T, xi):
        """
        Compute pressure given system information.
       
        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component
       
        Returns
        -------
        P : numpy.ndarray
            Array of pressure values [Pa] associated with each density and so equal in length
        """

        #logger = logging.getLogger(__name__)

        if len(xi) != len(self._nui):
            raise ValueError("Number of components in mole fraction list doesn't match components in nui. Check bead_config.")

        if T != self.T:
            self._temp_dependent_variables(T)

        self._xi_dependent_variables(xi)
        
        if np.isscalar(rho):
            rho = np.array([rho])
        elif type(rho) != np.ndarray:
            rho = np.array(rho)

        rho = rho*constants.Nav

        step = np.sqrt(np.finfo(float).eps) *rho * 10000.0
        # Decreasing step size by 2 orders of magnitude didn't reduce noise in P values
        nrho = np.size(rho)

        # computer rho+step and rho-step for better a bit better performance
        A = funcs.calc_A(np.append(rho + step, rho - step), xi, T, self._beads, self._beadlibrary, self._massi, self._nui, self._Cmol2seg, self._xsk, self._xskl, self._dkk, self._epsilonkl, self._sigmakl, self._dkl, self._l_akl, self._l_rkl, self._Ckl,self._x0kl, self._epsilonHB, self._Kklab, self._nk)

        P_tmp = (A[:nrho]-A[nrho:])*((constants.kb*T)/(2.0*step))*rho**2

        return P_tmp

    def fugacity_coefficient(self, P, rho, xi, T, dy=1e-4):

        """
        Compute fugacity coefficient.
      
        Parameters
        ----------
        P : float
            Pressure of the system [Pa]
        rho : float
            Molar density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component
    
        Returns
        -------
        mui : numpy.ndarray
            Array of chemical potential values for each component
        """

        #logger = logging.getLogger(__name__)

        if len(xi) != len(self._nui):
            raise ValueError("Number of components in mole fraction list doesn't match components in nui. Check bead_config.")

        if T != self.T:
            self._temp_dependent_variables(T)

        self._xi_dependent_variables(xi)

        Z = P / (rho * T * constants.Nav * constants.kb)
        phi_tmp = np.zeros(len(xi))

#        #### Traditional Central Difference Method
#        # Set step size in finite difference method
#        exp = np.floor(np.log10(rho))-3 # Make sure step size is three orders of magnitude lower
#        drho = 10**exp
#        logger.debug("    Compute phi for density, {}, with step size {}.".format(rho,drho))
#
#        # compute phi
#        Ares = funcs.calc_Ares(rho *  constants.Nav, xi, T, self._beads, self._beadlibrary, self._massi, self._nui, self._Cmol2seg, self._xsk, self._xskl, self._dkk, self._epsilonkl, self._sigmakl, self._dkl, self._l_akl, self._l_rkl, self._Ckl,self._x0kl, self._epsilonHB, self._Kklab, self._nk)
#        rhoi = rho*np.array(xi,float)
#        for i in range(np.size(phi_tmp)):
#            dAres = np.zeros(2)
#            for j, delta in enumerate((drho, -drho)):
#                rhoi_temp = np.copy(rhoi)
#                if rhoi_temp[i] != 0.:
#                    rhoi_temp[i] += delta
#                dAres[j] = self._calc_dAres_drhoi_wrap(T, rhoi_temp)
#            phi_tmp[i] = Ares + rho*(dAres[0] - dAres[1]) / (2.0 * drho) - np.log(Z) 
#            #with open("OldPhi.csv","a") as f:
#            #    f.write("{}, {}, {}, {}, {}, {}, {}\n".format(i,xi[i],phi_tmp[i], Ares, rho, dAres, drho))

        #### Transform y=log(rhoi) Central Difference Method without worrying about negative mole fractions 
        # Set step size in finite difference method
        y = np.log(rho*np.array(xi,float))
        #dy = 0.05

        # compute phi
        Ares = funcs.calc_Ares(rho *  constants.Nav, xi, T, self._beads, self._beadlibrary, self._massi, self._nui, self._Cmol2seg, self._xsk, self._xskl, self._dkk, self._epsilonkl, self._sigmakl, self._dkl, self._l_akl, self._l_rkl, self._Ckl,self._x0kl, self._epsilonHB, self._Kklab, self._nk)
        for i in range(np.size(phi_tmp)):
            if xi[i] != 0.0:
                dAres = np.zeros(2)
                for j, delta in enumerate((dy, -dy)):
                    y_temp = np.copy(y)
                    y_temp[i] += delta
                    dAres[j] = self._calc_dAres_drhoi_wrap(T, np.exp(y_temp))
                phi_tmp[i] = Ares + rho/np.exp(y[i])*(dAres[0] - dAres[1]) / (2.0 * dy) - np.log(Z)
            else:
                phi_tmp[i] = 0.0 # This should be zero, but to prevent the thermo calculation from complaining about diving by zero we give it a value, the mole fraction is zero though, so it'll go away.
        ##########################################

        phi = np.exp(phi_tmp)

        # Reset composition dependent variables
        self._xi_dependent_variables(xi)

        return phi

    def _calc_dAres_drhoi_wrap(self, T, rhoi):
        """
        Compute derivative of Helmholtz energy wrt to density.
      
        Parameters
        ----------
        T : float
            Temperature of the system [K]
        rhoi : float
            Molar density of each component, add up to the total density [mol/m^3]
    
        Returns
        -------
        Ares : float
            Helmholtz energy give number of moles, length of array rho
        """

        #logger = logging.getLogger(__name__)

        if T != self.T:
            self._temp_dependent_variables(T)

        # Calculate new xi values
        rho = np.array([np.sum(rhoi)])
        xi = rhoi/rho

        self._xi_dependent_variables(xi)

        Cmol2seg, xsk, xskl = funcs.calc_composition_dependent_variables(xi, self._nui, self._beads, self._beadlibrary)

        Ares = funcs.calc_Ares(rho * constants.Nav, xi, T, self._beads, self._beadlibrary, self._massi, self._nui, Cmol2seg, xsk, xskl, self._dkk, self._epsilonkl, self._sigmakl, self._dkl, self._l_akl, self._l_rkl, self._Ckl, self._x0kl, self._epsilonHB, self._Kklab, self._nk)

        return Ares

    def _chemicalpotential_old(self, P, rho, xi, T):
        """
        Compute chemical potential given system information.
      
        Parameters
        ----------
        P : float
            Pressure of the system [Pa]
        rho : float
            Molar density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component
    
        Returns
        -------
        mui : numpy.ndarray
            Array of chemical potential values for each component
        """

        logger = logging.getLogger(__name__)

        if len(xi) != len(self._nui):
            raise ValueError("Number of components in mole fraction list doesn't match components in nui. Check bead_config.")

        if T != self.T:
            self._temp_dependent_variables(T)

        self._xi_dependent_variables(xi)

        daresdxi = np.zeros_like(xi)
        mui = np.zeros_like(xi)

        # Set step size in finite difference method
        dnmol = 1.0E-4
        xi = np.array(xi,float)
        xi_tmp = xi[xi!=0.]
        if any(xi_tmp-dnmol < 0.):
            exp = np.floor(np.log10(min(xi_tmp)))-2 # Make sure step size is two orders of magnitude lower
            logger.debug("    Mole fraction, {}, is smaller than increment, {}. Use new increment, {}.".format(xi,dnmol,10**exp))
            dnmol = 10**exp

        # compute mui
        for i in range(np.size(mui)):
            dAres = np.zeros(2)
            ares = funcs.calc_Ares(rho * constants.Nav, xi, T, self._beads, self._beadlibrary, self._massi, self._nui, self._Cmol2seg, self._xsk, self._xskl,self._dkk, self._epsilonkl, self._sigmakl, self._dkl, self._l_akl, self._l_rkl, self._Ckl, self._x0kl, self._epsilonHB, self._Kklab, self._nk)
            for j, delta in enumerate((dnmol, -dnmol)):
                xi_temp = np.copy(xi)
                if xi_temp[i] != 0.:
                    xi_temp[i] += delta
                Cmol2seg_tmp, xsk_tmp, xskl_tmp = funcs.calc_composition_dependent_variables(xi_temp, self._nui, self._beads, self._beadlibrary)
                # xi_temp/=(nmol+delta)
                dAres[j] = funcs.calc_Ares(rho * constants.Nav, xi_temp, T, self._beads, self._beadlibrary, self._massi, self._nui, Cmol2seg_tmp, xsk_tmp, xskl_tmp, self._dkk, self._epsilonkl, self._sigmakl, self._dkl, self._l_akl, self._l_rkl, self._Ckl, self._x0kl, self._epsilonHB, self._Kklab, self._nk)
            daresdxi[i] = (dAres[0] - dAres[1]) / (2.0 * dnmol)

        # compute Z
        Z = P / (rho * T * constants.Nav * constants.kb)

        xjdaresdxj = np.sum(xi * daresdxi)
        for i in range(np.size(mui)):
            mui[i] = ares + Z - 1.0 + daresdxi[i] - xjdaresdxj - np.log(Z)
    
        return mui

    def density_max(self, xi, T, maxpack=0.65):

        """
        Estimate the maximum density based on the hard sphere packing fraction.
        
        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default: 0.65
            Maximum packing fraction
        
        Returns
        -------
        maxrho : float
            Maximum molar density [mol/m^3]
        """

        #logger = logging.getLogger(__name__)

        if T != self.T:
            self._temp_dependent_variables(T)

        self._xi_dependent_variables(xi)

        # estimate the maximum density based on the hard sphere packing fraction
        # etax, assuming a maximum packing fraction specified by maxpack
        maxrho = maxpack * 6.0 / (self._Cmol2seg * np.pi * np.sum(self._xskl * (self._dkl**3))) / constants.Nav

        return maxrho

    def param_guess(self, fit_params):
        """
        Generate initial guesses for the parameters to be fit.

        Parameters
        ----------
        fit_params : list[str]
            A list of parameters to be fit. See EOS mentation for supported parameter names. Cross interaction parameter names should be composed of parameter name and the other bead type, separated by an underscore (e.g. epsilon_CO2).

        Returns
        -------
        param_initial_guesses : numpy.ndarray, 
            An array of initial guesses for parameters, these will be optimized throughout the process.
    """

        #logger = logging.getLogger(__name__)

        l_fitparams = len(fit_params)
        if l_fitparams == 1:
            param_initial_guesses = np.array([l_fitparams])
        else:
            param_initial_guesses = np.zeros(1,l_fitparams)

        for i,param in enumerate(fit_params):
            if (param == "epsilon" or param.startswith('epsilon_')):
                param_initial_guesses[i] = 200.0
            elif param.startswith('epsilon'):
                param_initial_guesses[i] = 1000.0
            elif param.startswith('l_a'):
                param_initial_guesses[i] = 6.0
            elif param.startswith('l_r'):
                param_initial_guesses[i] = 12.0
            elif param.startswith('K'):
                param_initial_guesses[i] = 100.0e-30
            elif param.startswith('Sk'):
                param_initial_guesses[i] = 0.5
            else:
                raise ValueError("The parameter name {} does not fall under any of the catagories, epsilon, epsilon (assoc), l_a, l_r, K (assoc), or Sk".format(param)) 

        return param_initial_guesses

    def update_parameters(self, param_name, bead_names, param_value):
        r"""
        Update a single parameter value during parameter fitting process.

        To refresh those parameters that are dependent on to _beadlibrary or _crosslibrary, use method "parameter refresh".
        
        Parameters
        ----------
        param_name : str
            name of parameters being updated
        bead_names : list
            List of bead names to be changed. For a self interaction parameter, the length will be 1, for a cross interaction parameter, the length will be two.
        param_value : float
            Value of parameter
        """

        #logger = logging.getLogger(__name__)

        param_types = ["epsilon", "sigma", "l_r", "l_a", "Sk", "K"]

        if len(bead_names) > 2:
            raise ValueError("The bead names {} were given, but only a maximum of 2 are permitted.".format(", ".join(bead_names)))
        if not set(bead_names).issubset(self._beads):
            raise ValueError("The bead names {} were given, but they are not in the allowed list: {}".format(", ".join(bead_names),", ".join(self._beads)))

        # Non bonded parameters
        if (param_name in ["epsilon", "sigma", "l_r", "l_a", "Sk"]):
            # Self interaction parameter
            if len(bead_names) == 1:
                self._beadlibrary[bead_names[0]][param_name] = param_value
            # Cross interaction parameter
            elif len(bead_names) == 2:
                if bead_names[1] in list(self._crosslibrary.keys()):
                    if bead_names[0] in list(self._crosslibrary[bead_names[1]].keys()):
                        self._crosslibrary[bead_names[1]][bead_names[0]][param_name] = param_value
                    else:
                        self._crosslibrary[bead_names[1]] = {bead_names[0]:{param_name:param_value}}
                elif bead_names[0] in list(self._crosslibrary.keys()):
                    if bead_names[1] in list(self._crosslibrary[bead_names[0]].keys()): 
                        self._crosslibrary[bead_names[0]][bead_names[1]][param_name] = param_value
                    else:
                        self._crosslibrary[bead_names[0]] = {bead_names[1]:{param_name:param_value}}
                else:
                    self._crosslibrary[bead_names[0]] = {bead_names[1]:{param_name:param_value}}

        # Association Sites
        elif any([param_name.startswith('epsilon'), param_name.startswith('K')]):

            # Ensure sitenames are valid and on list
            if tmp[0] == True: tmp_name_full = param_name.replace("epsilon","")
            elif tmp[1] == True: tmp_name_full = param_name.replace("K","")
            flag = 0

            for site1 in self._sitenames:
                if tmp_name_full.startswith(site1):
                    tmp_name = tmp_name_full.replace(site1,"")
                    for site2 in self._sitenames:
                        if tmp_name == site2:
                            flag = 1
                            break
                    if flag == 1:
                        break
            if flag == 0:
                raise ValueError("site_names should be two different sites in the list: {}. You gave: {}".format(tmp_name_full,", ".join(sitenames=self._sitenames)))

            # Self interaction parameter
            if len(bead_names) == 1:
                self._beadlibrary[bead_names[0]][param_name+"".join(site_names)] = param_value
            # Cross interaction parameter
            elif len(bead_names) == 2:
                if bead_names[0] in list(self._crosslibrary[bead_names[1]].keys()):
                    self._crosslibrary[bead_names[1]][bead_names[0]][param_name+"".join(site_names)] = param_value
                else:
                    self._crosslibrary[bead_names[0]][bead_names[1]][param_name+"".join(site_names)] = param_value                        

        else:
            raise ValueError("The parameter name {} is not found in the allowed parameter types: {}".format(param_name,", ".join(param_types)))

    def parameter_refresh(self):
        r""" 
        To refresh dependent parameters
        
        Those parameters that are dependent on _beadlibrary and _crosslibrary attributes **must** be updated by running this function after all parameters from update_parameters method have been changed.
        """

        #logger = logging.getLogger(__name__)

        # Update Non bonded matrices
        self._epsilonkl, self._sigmakl, self._l_akl, self._l_rkl, self._Ckl = funcs.calc_interaction_matrices(self._beads, self._beadlibrary, crosslibrary=self._crosslibrary)

        # Update Association site matrices
        self._epsilonHB, self._Kklab, self._nk = funcs.calc_assoc_matrices(self._beads,self._beadlibrary,sitenames=self._sitenames,crosslibrary=self._crosslibrary)

        # Update temperature dependent variables
        if np.isnan(self.T) == False:
            self._temp_dependent_variables(self.T)


    def __str__(self):

        string = "Beads:" + str(self._beads) + "\n"
        string += "T:" + str(self.T) + "\n"
        return string





"""
DESPASITO
DESPASITO: Determining Equilibrium State and Parametrization: Application for SAFT, Intended for Thermodynamic Output
"""

# Add imports here
from .main import run

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions




import numpy as np
import logging

def reformat_ouput(cluster):
    r"""
    Takes a list of lists that combine lists and floats and reformats it into a 2D numpy array.
 
    Parameters
    ----------
    cluster : list[list[list/floats]]
        A list of lists, where the inner list is made up of lists and floats

    Returns
    -------
    matrix : numpy.ndarray
        A 2D matrix
    len_cluster : list
        a list of lengths for each of the columns (whether 1 for float, or len(list))
        
    """

    #logger = logging.getLogger(__name__)

    # Arrange data
    type_cluster = [type(x[0]) for x in cluster]

    # if input is a list or array
    if len(cluster) == 1:
        matrix = np.transpose(np.array(cluster[0]))
        if cluster[0][0] not in [list,np.ndarray,tuple]:
            len_cluster = [1]
        else:
            len_cluster = [len(cluster[0][0])]

    # If list of lists or arrays
    else:

        # Obtain dimensions of final matrix
        len_cluster = []
        for i,typ in enumerate(type_cluster):
            if typ in [list,np.ndarray,tuple]:
                len_cluster.append(len(cluster[i][0]))
            else:
                len_cluster.append(1)
        matrix_tmp = np.zeros([len(cluster[0]), sum(len_cluster)])

        # Transfer information to final matrix
        ind = 0
        for i, val in enumerate(cluster):
            matrix = np.transpose(np.array(val))
            l = len_cluster[i]
            if l == 1:
                matrix_tmp[:, ind] = np.array(matrix)
                ind += 1
            else:
                for j in range(l):
                    matrix_tmp[:, ind] = matrix[j]
                    ind += 1
        #matrix = np.transpose(np.array(matrix_tmp))
        matrix = np.array(matrix_tmp)

    return matrix, len_cluster

class BasinStep(object):
    r"""
    Custom basin step used by scipy.optimize.basinhopping function.
    
    """

    def __init__(self, stepmag, stepsize=0.05):
        r"""
            
        Parameters
        ----------
        stepmag : list
            List of step magnitudes
        stepsize : float
            default 0.05

        Attributes
        ----------
        stepmag : list
            List of step magnitudes
        stepsize : float
            default 0.05
            
        """

        #logger = logging.getLogger(__name__)
        
        self._stepsize = stepsize
        self._stepmag = stepmag

    def __call__(self, x):

        r"""
            
        Parameters
        ----------
        x : numpy.ndarray
            Guess in parameters values

        Returns
        -------
        basinstep : numpy.ndarray
            Suggested basin step used in scipy.optimize.basinhopping algorithm
            
        """

        logger = logging.getLogger(__name__)
        
        # Save intital guess in array
        xold = np.copy(x)

        # Loop for number of times to start over
        for j in range(1000):
            # reset array x
            x = np.copy(xold)
            breakloop = True
            # Iterate through array of step magnitudes
            for i, mag in enumerate(self._stepmag):
                # Add or subtract a random number within distribution of +- mag*stepsize
                x[i] += np.random.uniform(-mag * self._stepsize, mag * self._stepsize)
                # If a value of x is negative, don't  break the cycle
                if x[i] < 0.0:
                    breakloop = False
            if breakloop: break
            logger.info("Basin Step after {} iterations:\n    {}".format(j,x))
        return x

class BasinBounds(object):
    r"""
    Object used by scipy.optimize.basinhopping to set bounds of parameters.
    """

    def __init__(self, bounds):
        r"""
            
        Parameters
        ----------
        bounds : numpy.ndarray
            Bounds on parameters
    
        Attributes
        ----------
        xmin : numpy.ndarray
            Array of minimun values for each parameter
        xman : numpy.ndarray
            Array of maximum values for each parameter
            
        """
        bounds = np.transpose(np.array(bounds))

        self.xmin = bounds[0]
        self.xmax = bounds[1]

    def __call__(self, **kwargs):
        r"""
            
        Parameters
        ----------
        x_new : numpy.ndarray
            Guess in parameters values

        Returns
        -------
        value : bool
            A true or false value that says whether the guess in parameter value is within bounds
            
        """
        x = kwargs["x_new"]
        tmax = bool(np.all(x <= self.xmax))
        tmin = bool(np.all(x >= self.xmin))
        return tmax and tmin


def compute_SAFT_obj(beadparams, opt_params, eos, exp_dict):
    r"""
    Fit defined parameters for equation of state object with given experimental data. 

    Each set of experimental data is converted to an object with the build in ability to evaluate its part of objective function. 
    To add another type of supported experimental data, add a class to the fit_classes.py file.

    Parameters
    ----------
    beadparams0 : numpy.ndarray, 
        An array of initial guesses for parameters, these will be optimized throughout the process.
    opt_params : dict
        Parameters used in basin fitting algorithm

        - fit_bead (str) - Name of bead whose parameters are being fit, should be in bead list of beadconfig
        - fit_params (list[str]) - This list of contains the name of the parameter being fit (e.g. epsilon). See EOS documentation for supported parameter names. Cross interaction parameter names should be composed of parameter name and the other bead type, separated by an underscore (e.g. epsilon_CO2).

    eos : obj
        Equation of state output that writes pressure, max density, chemical potential, updates parameters, and evaluates objective functions. For parameter fitting algorithm See equation of state documentation for more details.
    exp_dict : dict
        Dictionary of experimental data objects.

    Returns
    -------
    Objective : float
        Value of sum of objective values according to appropriate weights. Output file saved in current working directory.
        
    """

    logger = logging.getLogger(__name__)

    # Update beadlibrary with test paramters
    for i, param in enumerate(opt_params['fit_params']):
        fit_params_list = param.split("_")
        if len(fit_params_list) == 1:
            eos.update_parameters(fit_params_list[0], [opt_params['fit_bead']], beadparams[i])
        elif len(fit_params_list) == 2:
            eos.update_parameters(fit_params_list[0], [opt_params['fit_bead'], fit_params_list[1]], beadparams[i])
        else:
            raise ValueError("Parameters for only one bead are allowed to be fit at one time. Please only list one bead type in your fit parameter name.")
    eos.parameter_refresh()

    # Compute obj_function
    obj_function = []
    for key,data_obj in exp_dict.items():
        try:
            obj_function.append(data_obj.objective(eos))
        except:
            raise ValueError("Failed to evaluate objective function for {} of type {}.".format(key,data_obj.name))

    # Write out parameters and objective functions for each dataset
    logger.info("\nParameters: {}\nValues: {}\nExp. Data: {}\nObj. Values: {}\nTotal Obj. Value: {}".format(opt_params['fit_params'],beadparams,list(exp_dict.keys()),obj_function,sum(obj_function)))

    return sum(obj_function)

